(*********************************************************************

  aux

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)



Require Export LibTactics LibReflect LibRelation LibNat LibVar LibList LibListSorted LibFset Morphisms LibListExt LibFsetExt Utf8.

Fixpoint lt_dec (m n : nat) :=
  match m, n with
    | O, O => false
    | O, S _ => true
    | S _, O => false
    | S m', S n' => lt_dec m' n'
  end.
                      
Instance lt_decidable : forall (m n : nat), Decidable (m<n).
Proof. intros. applys~ decidable_make (lt_dec m n).
  gen n. induction m; intros; destruct n; simpl.
  symmetry. apply~ isTrue_false. nat_math.
  symmetry. apply~ isTrue_true. nat_math.
  symmetry. apply~ isTrue_false. nat_math.
  asserts_rewrite ((S m < S n) = (m < n)).
    extens. iff; nat_math.
  auto.  
Qed.

Lemma min_same: forall n, (min n n) = n.
Proof.
  induction n; simpls~.
Qed.

Hint Resolve min_same.

Definition max (m n : nat) := ifb (m < n) then n else m.

Lemma max0 : forall m n, max m n = 0 -> m = 0 /\ n = 0.
Proof.
  introv H. unfolds in H; cases_if~; nat_math.
Qed.

Lemma max0neutrall : forall m, max 0 m = m.
Proof.
  introv. unfolds. cases_if; nat_math.
Qed.

Lemma max0neutralr : forall m, max m 0 = m.
Proof.
  introv. unfolds. cases_if; nat_math.
Qed.

Lemma maxcomm : forall m n, max m n = max n m.
Proof.
  introv. unfolds; cases_if; cases_if~; nat_math.
Qed.

Lemma maxassoc : forall m n p, max m (max n p) = max (max m n) p.
Proof.
  introv. unfolds; repeat cases_if~; nat_math.
Qed.
