(*********************************************************************

  HOC27BarbedRevisited

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)

Require Import HOC01Defs.
Require Import HOC02BaseLemmas.
Require Import HOC03CanonicalLemmas.
Require Import HOC04DefLTS.
Require Import HOC05WFProcesses.
Require Import HOC06FreshLemmas.
Require Import HOC07SizeLemmas.
Require Import HOC08SubstLemmas.
Require Import HOC09GVLemmas.
Require Import HOC10CongrLemmas.
Require Import HOC11TransLemmas.
Require Import HOC12Guarded.
Require Import HOC13Bisimulations.
Require Import HOC14IObis.
Require Import HOC17NORbis.
Require Import HOC19Coincide.
Require Import HOC20Barbed.
Require Import HOC23PrimeDecomposition.
Require Import HOC24PreCompleteness.
Require Import HOC25Axiomatization.
Require Import HOC26EarlyBisimulation.

Require Import Relations.

(* Connecting with congruence *)

Lemma congr_context_closed : context_closed wfcgr.
Proof with automate.
  introv Hc. gen p q. inductions C; simpls~.

  + applys~ wfcgr_send. 
  + applys~ wfcgr_abs. 
  + applys~ wfcgr_par. reflexivity.
  + applys~ wfcgr_par. reflexivity. 
Qed.

Lemma congr_out_barb_preserving : out_barb_preserving wfcgr.
Proof with automate.
  introv Hc. inductions Hc; jauto. inductions H; jauto.

  introv (p' & p'' & Htp). inverts Htp...
  exists wfp_Nil q. constructor.

  introv (p' & p'' & Htp). inverts Htp.

  introv (r' & r'' & Htpar). inverts Htpar...
    destruct ap; inverts H5. 
    lets (p1 & p2 & Htp') : IHwfcgr_step1 a.
    eexists. eexists. eassumption.
    eexists. eexists. apply act1_out. eassumption.
    destruct aq; inverts H5. 
    lets (p1 & p2 & Htp') : IHwfcgr_step2 a.
    eexists. eexists. eassumption.
    eexists. eexists. apply act2_out. eassumption.

  introv (p' & p'' & Htp).
    eexists. eexists. apply act1_out. eassumption.

  introv (p' & p'' & Htp). inverts Htp...
    destruct ap; inverts H3. eexists. eexists. eassumption.
    inverts H2.

  introv (p' & p'' & Htp). inverts Htp...
    destruct ap; inverts H3. eexists. eexists. apply act2_out. eassumption.
    destruct aq; inverts H3. eexists. eexists. apply act1_out. eassumption.

  introv (p' & p'' & Htp). inverts Htp...
    destruct ap; inverts H3. eexists. eexists. repeat apply act1_out. eassumption.
    destruct aq; inverts H3. inverts H2... 
      destruct ap; inverts H4. 
        eexists. eexists. apply act1_out. apply act2_out. eassumption.
      destruct aq; inverts H4.
        eexists. eexists. repeat apply act2_out. eassumption.

  introv (p' & p'' & Htp). inverts Htp...
    destruct ap; inverts H3. inverts H2... 
      destruct ap; inverts H4. 
        eexists. eexists. repeat apply act1_out. eassumption.
      destruct aq; inverts H4.
        eexists. eexists. apply act2_out. apply act1_out. eassumption.
    destruct aq; inverts H3.  
        eexists. eexists. repeat apply act2_out. eassumption.   
Qed.

Lemma congr_async_barbed_relation : async_barbed_relation wfcgr.
Proof.
  splits~.
    apply wfcgr_is_eq_rel.
    apply congr_tau.
    apply congr_context_closed.
    apply congr_out_barb_preserving.
Qed.

Lemma congr_async_barbed : forall p q, p ≡* q -> async_barbed p q.
Proof.
  introv Hc. exists wfcgr. splits*.
  apply congr_async_barbed_relation.
Qed.

Instance async_barbed_refl : Reflexive async_barbed.
Proof.
  introv; apply congr_async_barbed. reflexivity.
Qed.

Hint Resolve async_barbed_refl async_barbed_sym async_barbed_trans.

Lemma async_barbed_congr_nil_l : forall p, async_barbed p (wfp_Par p wfp_Nil).
Proof.
  introv; apply congr_async_barbed. apply wfcgr_nil_l. 
Qed.

Lemma async_barbed_congr_nil_r : forall p, async_barbed (wfp_Par p wfp_Nil) p.
Proof.
  introv; apply congr_async_barbed. apply wfcgr_nil_r.
Qed.

Lemma async_barbed_congr_par_com : forall p q, async_barbed (wfp_Par p q) (wfp_Par q p).
Proof.
  introv; apply congr_async_barbed. apply wfcgr_par_com.
Qed.

Lemma async_barbed_congr_par_assoc_l : forall p q r, async_barbed (wfp_Par p (wfp_Par q r)) (wfp_Par (wfp_Par p q) r).
Proof.
  introv; apply congr_async_barbed. apply wfcgr_par_assoc_l. 
Qed.

Lemma async_barbed_congr_par_assoc_r : forall p q r, async_barbed (wfp_Par (wfp_Par p q) r) (wfp_Par p (wfp_Par q r)).
Proof.
  introv; apply congr_async_barbed. apply wfcgr_par_assoc_r.
Qed.

Lemma async_barbed_congr_send : forall p q a, 
  async_barbed p q -> async_barbed (wfp_Send a p) (wfp_Send a q).
Proof.
  introv Hab. 
  set (Ctx := CSend a CHole).
  apply async_barbed_context_closed in Hab. specializes Hab Ctx.
  subst Ctx; simpls~.
Qed.

Lemma async_barbed_congr_abs : forall p q a X, 
  async_barbed p q -> async_barbed (wfp_Abs a X p) (wfp_Abs a X q).
Proof.
  introv Hab. 
  set (Ctx := CAbs a X CHole).
  apply async_barbed_context_closed in Hab. specializes Hab Ctx.
  subst Ctx; simpls~.
Qed.

Lemma async_barbed_congr_par_l : forall p q r, 
  async_barbed p q -> async_barbed (wfp_Par p r) (wfp_Par q r).
Proof.
  introv Hab. 
  set (Ctx := CParL CHole r).
  apply async_barbed_context_closed in Hab. specializes Hab Ctx.
  subst Ctx; simpls~.
Qed.

Lemma async_barbed_congr_par_r : forall p q r, 
  async_barbed p q -> async_barbed (wfp_Par r p) (wfp_Par r q).
Proof.
  introv Hab. 
  set (Ctx := CParR r CHole).
  apply async_barbed_context_closed in Hab. specializes Hab Ctx.
  subst Ctx; simpls~.
Qed.

Lemma async_barbed_congr_par : forall p q r s,
  async_barbed p q -> async_barbed r s -> 
    async_barbed (wfp_Par p r) (wfp_Par q s).
Proof.
  introv Hab_pq Hab_rs.
  apply async_barbed_context_closed in Hab_pq. specializes Hab_pq (CParL CHole r).
  simpls. etransitivity. eassumption.
  apply async_barbed_context_closed in Hab_rs. specializes Hab_rs (CParR q CHole).
  simpls~.
Qed.

Lemma async_barbed_congr_send_inv : forall p q a,
  async_barbed (wfp_Send a p) (wfp_Send a q) -> async_barbed p q.
Proof with automate.
  introv Hab. lets (X & _) : find_fresh (@nil process).
  
  lets Hab_cc : async_barbed_context_closed Hab (CParL CHole (wfp_Abs a X (wfp_Gvar X))); simpls.
  
  lets Hin_p : TrIn a X (wfp_Gvar X).
  lets Hout_p : TrOut a p.
  lets Htau_p : TrTau1 Hout_p Hin_p. simpls... cases_if*.
  
  lets (q' & Htau_q & Hab') : async_barbed_tau Hab_cc Htau_p.
  
  inverts Htau_q; subst_wfp;
  assert (p0 = wfp_Send a q) by rewrite_wfp_equal;
  assert (q0 = wfp_Abs a X (wfp_Gvar X)) by rewrite_wfp_equal; substs;
  try solve [inverts H2].

  clear H H0. inverts H2...
  assert (wfp_inst_Abs fq q = q).
    lets Hyp : wfp_trans_in_abs_eq H3.
    inverts Hyp as Hyp. simpls... cases_if*.
    destruct Hyp as (Y & HneqY & HfY & Heq); subst.
    simpls. rewrite~ <- wfp_subst_decomposition_g... cases_if*.
  rewrite H in *.

  etransitivity. apply async_barbed_congr_nil_l.
  etransitivity. apply async_barbed_congr_par_com.
  etransitivity. eassumption.
  etransitivity. apply async_barbed_congr_par_com.
  etransitivity. apply async_barbed_congr_nil_r.
  reflexivity.
Qed.

Lemma async_barbed_congr_abs_inv : forall p q a X,
  async_barbed (wfp_Abs a X p) (wfp_Abs a X q) -> async_barbed p q.
Proof with automate.
  introv Hab. 
 
  lets Hab_cc : async_barbed_context_closed Hab (CParL CHole (wfp_Send a (wfp_Gvar X))); simpls.
  
  lets Hin_p : TrIn a X p.
  lets Hout_p : TrOut a (wfp_Gvar X).
  lets Htau_p : TrTau2 Hin_p Hout_p. simpls... rewrite wfp_subst_gvar_idem in *. 
  
  lets (q' & Htau_q & Hab') : async_barbed_tau Hab_cc Htau_p.
  
  inverts Htau_q; subst_wfp;
  assert (q0 = wfp_Send a (wfp_Gvar X)) by rewrite_wfp_equal; substs;
  try solve [inverts H2].

  clear H0. inverts H3... 
  assert (q'' = wfp_Gvar X) by rewrite_wfp_equal; substs; clear H1.
  assert (wfp_inst_Abs fp (wfp_Gvar X) = q).
    lets Hyp : wfp_trans_in_abs_eq H2.
    inverts Hyp as Hyp. simpls... rewrite~ wfp_subst_gvar_idem. 
    destruct Hyp as (Y & HneqY & HfY & Heq); subst.
    simpls. rewrite~ <- wfp_subst_decomposition_g... rewrite~ wfp_subst_gvar_idem.
  rewrite H in *.

  etransitivity. apply async_barbed_congr_nil_l.
  etransitivity. eassumption.
  etransitivity. apply async_barbed_congr_nil_r.
  reflexivity.
Qed.
    
(* Swapping *)

Lemma async_barbed_swap_fresh : forall (p q : wfprocess) X Y, 
  Y # p -> Y # q -> async_barbed p q ->
    async_barbed (wfp_swap X Y p) (wfp_swap X Y q).
Proof with automate.
  introv Hfyp Hfyq Hab.
  tests Heq : (X = Y). repeat rewrite~ wfp_swap_equal.
 
  lets (a & _) : find_fresh_chan (@nil process).
  lets Hab_cc_xy : async_barbed_context_closed Hab.
  specializes Hab_cc_xy (CParL (CAbs a X CHole) (wfp_Send a (wfp_Gvar Y))); simpls.
 
  lets Hin_p : TrIn a X p.
  lets Hout_p : TrOut a (wfp_Gvar Y).
  lets Htau_p : TrTau2 Hin_p Hout_p. simpls.
  rewrite wfp_gfresh_subst_is_swap in Htau_p; auto.
  etransitivity. apply async_barbed_congr_nil_l.

  lets (q' & Htau_q & Hab') : async_barbed_tau Hab_cc_xy Htau_p.
  etransitivity. eassumption.

  inverts Htau_q; subst_wfp;
  assert (q0 = wfp_Send a (wfp_Gvar Y)) by rewrite_wfp_equal; substs; 
  try solve [inverts H2].

  clear H0. inverts H3...
  assert (q'' = wfp_Gvar Y) by rewrite_wfp_equal; substs. clear H1.
  assert (wfp_inst_Abs fp (wfp_Gvar Y) = wfp_swap X Y q).
    rewrite~ <- wfp_gfresh_subst_is_swap.
    lets Hyp : wfp_trans_in_abs_eq H2. inverts Hyp as Hyp.
    simpls~. destruct Hyp as (Z & HneqZ & HfZ & Heq'). subst.
    simpls. rewrite~ <- wfp_subst_decomposition_g.
  rewrite H in *. apply async_barbed_congr_nil_r.  
Qed.

Lemma async_barbed_swap : forall p q X Y, async_barbed p q ->
  async_barbed (wfp_swap X Y p) (wfp_swap X Y q).
Proof with automate.
  introv Hab. tests Heq : (X = Y).
  repeat rewrite~ wfp_swap_equal.
  lets (Z & Hfz) : find_fresh (proc p :: proc q :: proc (wfp_Gvar X) :: proc (wfp_Gvar Y) :: nil).
  rewrite Forall_to_conj_4 in Hfz; destruct Hfz as (Hfzp & Hfzq & Hfzx & Hfzy).
  simpls; rewrite gfresh_gvar in *.
  
  (* Swap X and Z *)
  lets Hab_xz : async_barbed_swap_fresh X Z Hfzp Hfzq Hab.
  lets Hab_zy : async_barbed_swap_fresh Y X Hab_xz.
    rewrite~ <- wfp_gfresh_subst_is_swap; simpls; fresh_solve.
    rewrite~ <- wfp_gfresh_subst_is_swap; simpls; fresh_solve.
  lets Hab_yx : async_barbed_swap_fresh Z Y Hab_zy.
    simpls; repeat (rewrite~ <- gfresh_subst_is_swap; simpls; fresh_solve).
    simpls; repeat (rewrite~ <- gfresh_subst_is_swap; simpls; fresh_solve).

  rewrite <- wfp_gfresh_subst_is_swap in Hab_yx;
    try (simpls; repeat (rewrite~ <- gfresh_subst_is_swap; simpls; fresh_solve)).

  rewrite wfp_subst_on_swap in Hab_yx... cases_if*.
  rewrite wfp_swap_sym in Hab_yx. unfolds swap_gvar; repeat cases_if*.
  rewrite wfp_subst_on_swap in Hab_yx... cases_if*.
  unfolds swap_gvar; repeat cases_if*.
  rewrite~ wfp_gfresh_subst_is_swap in Hab_yx. 
  rewrite wfp_swap_involutive in Hab_yx.
  etransitivity. eassumption.

  rewrite <- wfp_gfresh_subst_is_swap;
    try (simpls; repeat (rewrite~ <- gfresh_subst_is_swap; simpls; fresh_solve)).
  rewrite wfp_subst_on_swap... cases_if*.
  rewrite wfp_swap_sym. unfolds swap_gvar; repeat cases_if*.
  rewrite wfp_subst_on_swap... cases_if*.
  unfolds swap_gvar; repeat cases_if*.
  rewrite~ wfp_gfresh_subst_is_swap. 
  rewrite~ wfp_swap_involutive.
Qed.

(* Connecting with the required bisimulations *)

Lemma async_barbed_out_normal_early : out_normal_relation_early async_barbed.
Proof with automate.
  introv Hr Ht Hfmp Hfmq Hfxp Hfxq.

  lets (n & Hcn) : find_fresh_chan (proc p :: proc q :: Send m Nil :: nil).
  rewrite Forall_to_conj_3 in Hcn. destruct Hcn as (Hfnp & Hfnq & Hfnm).
  rewrite cfresh_send in Hfnm. destruct Hfnm as (_ & Hfnm).
  assert (Hfxp'' : X # p'').
    lets Hgv : GV_trans_out Ht.
    intros Hin; apply Hfxp.
    rewrite Hgv; rewrite~ in_union.

  assert (Hfma : m <> a).
    eapply wfp_cfresh_out with (m := m) in Ht; jauto.

  assert (Hfna : n <> a).
    eapply wfp_cfresh_out with (m := n) in Ht; jauto.

  lets (Y & Hfy) : find_fresh (Gvar X :: nil).
    inverts Hfy as Hfy. clear H3. rewrite gfresh_gvar in Hfy.

  set (Ctx := CParL CHole (wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil))))).    
  lets Hcc : async_barbed_context_closed Hr Ctx. subst Ctx. simpls. 

  (* Left side transitions *)
  lets Hinl : TrIn a Y ((wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
  lets Htaul : TrTau1 Ht Hinl. simpls...
  repeat (rewrite wfp_eq_subst_abs2 in Htaul; jauto)... cases_if*.
  
  lets (Q & Htaur & HPQ) : async_barbed_tau Hcc Htaul.

  set (P := (wfp_Par p'
             (wfp_Par (wfp_Abs m X p'')
                (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))));
    fold P in HPQ.

  assert (Hout_P : perform_output P n).
  eexists. eexists. subst. repeat applys~ act2_out. 

  lets Hout_Q : async_barbed_out_barb_preserving HPQ Hout_P.

  inverts Htaur...

  + destruct ap; inverts H3. 
    assert (q0 = wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst.
      destruct Hout_Q as (q' & q'' & Htr).
      inverts Htr...
      - false. eapply (false_trans_out n w); try eassumption.
        eapply wfp_cfresh_tau in H2. apply H2. auto.
      - assert (q0 = wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
          rewrite_wfp_equal. subst q0. inverts H4.
  
  + assert (q0 = wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0. inverts H2.

  + assert (q0 = wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0.

    lets Heq : wfp_trans_in_chan_eq H3. subst a0.
    rename p''0 into q'', p'0 into q'. exists q' q''; splits*.
    assert (Hfxq'' : X # q'').
      lets Hgv : GV_trans_out H2.
      intros Hin; apply Hfxq.
      rewrite Hgv; rewrite~ in_union.
    assert (wfp_inst_Abs fq q'' = (wfp_Par (wfp_Abs m X q'')
            (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      lets Heq : wfp_trans_in_abs_eq H3.
      inverts Heq as Heq. simpls...
      repeat (rewrite wfp_eq_subst_abs2; jauto)... cases_if*.
      destruct Heq as (Z & HneqZ & HfZ & Heq); subst.
      simpl; subst_wfp. rewrite~ <- wfp_subst_decomposition_g...
      repeat (rewrite wfp_eq_subst_abs2; jauto)... cases_if*.
    repeat rewrite H in *; clear H.

    assert (Htau_P : {{P -- Tau ->> AP (wfp_Par p' (wfp_Par (wfp_Abs m X p'') (wfp_Par (wfp_Nil) (wfp_Nil))))}}).      
      subst P. do 2 apply act2_tau. 
      lets Htau : TrTau2 n (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil) wfp_Nil wfp_Nil. specializes Htau (AbsPure X wfp_Nil). simpls...

    lets (Q' & Htau_Q & HP'Q') : async_barbed_tau HPQ Htau_P.

    assert (Hno_out_P' : ~ perform_output (wfp_Par p' (wfp_Par (wfp_Abs m X p'') (wfp_Par wfp_Nil wfp_Nil))) n).
      introv Htr. destruct Htr as (P' & P'' & Htr).
      inverts Htr... apply false_trans_out in H5; auto. 
        apply (wfp_cfresh_out _ _ _ _ _ Hfnp Ht).
        assert (q0 = wfp_Par (wfp_Abs m X p'') (wfp_Par wfp_Nil wfp_Nil)).
          rewrite_wfp_equal. subst. clear H4. destruct aq; inverts H6...
        inverts H5... inverts H6. inverts H6; subst_wfp; inverts H8.

    assert (Hno_out_Q' : ~ perform_output Q' n).
      introv Hout. elim Hno_out_P'. eapply async_barbed_out_barb_preserving.
      apply async_barbed_sym. eassumption. auto.

    inverts Htau_Q; subst_wfp; 
    assert (q0 = (wfp_Par (wfp_Abs m X q'') (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))) by rewrite_wfp_equal; subst...
    
    - destruct ap; inverts H6.
      false. apply Hno_out_Q'.
        eexists. eexists. repeat applys~ act2_out.

    - clear H4. destruct aq; inverts H6...
      inverts H5; subst_wfp;
      assert (q0 = (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil))) by rewrite_wfp_equal; subst... 
      * inverts H6.
      * clear H4. destruct aq; inverts H7. inverts H6; subst_wfp;
        assert (q0 = wfp_Send n wfp_Nil) by rewrite_wfp_equal; subst; subst_wfp;
        try solve [inverts H5]. inverts H7...

        assert (Heq : wfp_inst_Abs fp wfp_Nil = wfp_Nil).
          lets Heq : wfp_trans_in_abs_eq H5.
          inverts Heq as Heq. simpls...
          destruct Heq as (Z & HneqZ & HfZ & Heq); subst.
          simpl; subst_wfp. rewrite~ <- wfp_subst_decomposition_g...
        repeat rewrite Heq in *; clear Heq.
  
        (* CONGRUENCE *)
        etransitivity. apply async_barbed_congr_par_com.
        etransitivity. Focus 2. apply async_barbed_congr_par_com.
        etransitivity. apply async_barbed_congr_nil_l.
        etransitivity. apply async_barbed_congr_nil_l.
        etransitivity. apply async_barbed_congr_par_assoc_r.
        etransitivity. apply async_barbed_congr_par_assoc_r.
        etransitivity. eassumption.
        etransitivity. apply async_barbed_congr_par_assoc_l.
        etransitivity. apply async_barbed_congr_par_assoc_l.
        etransitivity. apply async_barbed_congr_nil_r.
        etransitivity. apply async_barbed_congr_nil_r.
        reflexivity.     
        
      * inverts H6...
      * lets Heq : wfp_trans_in_chan_eq H6. subst a0.
        inverts H7...
          inverts H8. assert (q0 = wfp_Send n wfp_Nil).
            rewrite_wfp_equal. subst. inverts* H8. false~ Hfnm.
    - clear H1. inverts H6; subst_wfp; assert (q0 = wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)) by rewrite_wfp_equal; subst...
      * clear H4; destruct ap; inverts H8... 
        inverts H7. simpls. false Hno_out_Q'.
        eexists. eexists. repeat applys~ act2_out.
      * clear H4; destruct aq; inverts H8...
        inverts H7; subst_wfp; 
        assert (q0 = wfp_Send n wfp_Nil) by rewrite_wfp_equal; subst.
        inverts H6... inverts H8. false Hno_out_Q'.
        eexists. eexists. repeat applys~ act2_out. 
        false. rewrite wfprocess_eq in H; simpls; inverts H.
        inverts H6.

    - clear H1. inverts H6... inverts H7.
      assert (q0 = wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)).
        rewrite_wfp_equal. subst; clear H4. destruct aq; inverts H8.
      inverts H7... inverts H6. assert (q0 = wfp_Send n wfp_Nil).
        rewrite_wfp_equal. subst. clear H4. inverts H6...
      false. eapply (false_trans_in n). Focus 2. apply H5.
        apply (wfp_cfresh_out _ _ _ _ _ Hfnq H2).     

  + assert (q0 = wfp_Abs a Y (wfp_Par (wfp_Abs m X (wfp_Gvar Y)) (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0. inverts H3.
Qed.

Lemma async_barbed_out_1 : forall (p q : wfprocess), (async_barbed p q) -> 
    forall (a : chan) (p' p'' : wfprocess), 
    {{p -- LabOut a p'' ->> (AP p')}} ->
      exists (q' q'' : wfprocess),
        ({{q -- LabOut a q'' ->> (AP q')}} /\ (async_barbed p' q')).
Proof with automate.
  introv Hr Ht.

  lets (X & Hfx) : find_fresh (proc p :: proc q :: nil).
  rewrite Forall_to_conj_2 in Hfx; destruct Hfx as (Hfxp & Hfxq).

  assert (Hfxp'' : X # p'').
    lets Hgv : GV_trans_out Ht.
    intros Hin; apply Hfxp.
    rewrite Hgv; rewrite~ in_union.

  lets (n & Hcn) : find_fresh_chan (proc p :: proc q :: nil).
  rewrite Forall_to_conj_2 in Hcn. destruct Hcn as (Hfnp & Hfnq).

  assert (Hfna : n <> a).
    eapply wfp_cfresh_out with (m := n) in Ht; jauto.

  lets (Y & Hfy) : find_fresh (Gvar X :: nil).
    inverts Hfy as Hfy. clear H3. rewrite gfresh_gvar in Hfy.

  set (Ctx := CParL CHole (wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil))))).  
  
  lets Hcc : async_barbed_context_closed Hr Ctx. subst Ctx. simpls. 

  (* Left side transitions *)
  lets Hinl : TrIn a Y ((wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
  lets Htaul : TrTau1 Ht Hinl. simpls...
  repeat (rewrite wfp_eq_subst_abs2 in Htaul; jauto)... 

  lets (Q & Htaur & HPQ) : async_barbed_tau Hcc Htaul.

  set (P := (wfp_Par p'
             (wfp_Par wfp_Nil
                (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))));
    fold P in HPQ.

  assert (Hout_P : perform_output P n).
  eexists. eexists. subst. repeat applys~ act2_out. 

  lets Hout_Q : async_barbed_out_barb_preserving HPQ Hout_P.

  inverts Htaur...

  + destruct ap; inverts H3. 
    assert (q0 = wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst.
      destruct Hout_Q as (q' & q'' & Htr).
      inverts Htr...
      - false. eapply (false_trans_out n w); try eassumption.
        eapply wfp_cfresh_tau in H2. apply H2. auto.
      - assert (q0 = wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
          rewrite_wfp_equal. subst q0. inverts H4.
  
  + assert (q0 = wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0. inverts H2.

  + assert (q0 = wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0.

    lets Heq : wfp_trans_in_chan_eq H3. subst a0.
    rename p''0 into q'', p'0 into q'. exists q' q''; splits*.
    assert (Hfxq'' : X # q'').
      lets Hgv : GV_trans_out H2.
      intros Hin; apply Hfxq.
      rewrite Hgv; rewrite~ in_union.
    assert (wfp_inst_Abs fq q'' = (wfp_Par wfp_Nil
            (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      lets Heq : wfp_trans_in_abs_eq H3.
      inverts Heq as Heq. simpls...
      repeat (rewrite wfp_eq_subst_abs2; jauto)... 
      destruct Heq as (Z & HneqZ & HfZ & Heq); subst.
      simpl; subst_wfp. rewrite~ <- wfp_subst_decomposition_g...
      repeat (rewrite wfp_eq_subst_abs2; jauto)... 
    repeat rewrite H in *; clear H.

    assert (Htau_P : {{P -- Tau ->> AP (wfp_Par p' (wfp_Par wfp_Nil (wfp_Par (wfp_Nil) (wfp_Nil))))}}).      
      subst P. do 2 apply act2_tau. 
      lets Htau : TrTau2 n (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil) wfp_Nil wfp_Nil. specializes Htau (AbsPure X wfp_Nil). simpls...

    lets (Q' & Htau_Q & HP'Q') : async_barbed_tau HPQ Htau_P.

    assert (Hno_out_P' : ~ perform_output (wfp_Par p' (wfp_Par wfp_Nil (wfp_Par wfp_Nil wfp_Nil))) n).
      introv Htr. destruct Htr as (P' & P'' & Htr).
      inverts Htr... apply false_trans_out in H5; auto. 
        apply (wfp_cfresh_out _ _ _ _ _ Hfnp Ht).
        assert (q0 = wfp_Par wfp_Nil (wfp_Par wfp_Nil wfp_Nil)).
          rewrite_wfp_equal. subst. clear H4. destruct aq; inverts H6...
        inverts H5... inverts H6. inverts H6; subst_wfp; inverts H8.

    assert (Hno_out_Q' : ~ perform_output Q' n).
      introv Hout. elim Hno_out_P'. eapply async_barbed_out_barb_preserving.
      apply async_barbed_sym. eassumption. auto.

    inverts Htau_Q; subst_wfp; 
    assert (q0 = (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))) by rewrite_wfp_equal; subst...
    
    - destruct ap; inverts H6.
      false. apply Hno_out_Q'.
        eexists. eexists. repeat applys~ act2_out.

    - clear H4. destruct aq; inverts H6...
      inverts H5; subst_wfp;
      assert (q0 = (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil))) by rewrite_wfp_equal; subst... 
      * inverts H6.
      * clear H4. destruct aq; inverts H7. inverts H6; subst_wfp;
        assert (q0 = wfp_Send n wfp_Nil) by rewrite_wfp_equal; subst; subst_wfp;
        try solve [inverts H5]. inverts H7...

        assert (Heq : wfp_inst_Abs fp wfp_Nil = wfp_Nil).
          lets Heq : wfp_trans_in_abs_eq H5.
          inverts Heq as Heq. simpls...
          destruct Heq as (Z & HneqZ & HfZ & Heq); subst.
          simpl; subst_wfp. rewrite~ <- wfp_subst_decomposition_g...
        repeat rewrite Heq in *; clear Heq.
  
        (* CONGRUENCE *)
        etransitivity. apply async_barbed_congr_nil_l.
        etransitivity. Focus 2. apply async_barbed_congr_nil_r.
        etransitivity. apply async_barbed_congr_nil_l.
        etransitivity. apply async_barbed_congr_nil_l.
        etransitivity. apply async_barbed_congr_par_assoc_r.
        etransitivity. apply async_barbed_congr_par_assoc_r.
        etransitivity. eassumption.
        etransitivity. apply async_barbed_congr_par_assoc_l.
        etransitivity. apply async_barbed_congr_par_assoc_l.
        etransitivity. apply async_barbed_congr_nil_r.
        etransitivity. apply async_barbed_congr_nil_r.
        reflexivity.     
        
      * inverts H6...
      * inverts H6... 

    - clear H1. inverts H6... inverts H7.
      assert (q0 = wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)).
        rewrite_wfp_equal. subst; clear H4. destruct aq; inverts H8.
      inverts H7... assert (q0 = wfp_Send n wfp_Nil).
        rewrite_wfp_equal. subst. clear H4. inverts H6...
      false. eapply (false_trans_out n). Focus 2. apply H5.
        apply (wfp_cfresh_out _ _ _ _ _ Hfnq H2).     
      assert (q0 = wfp_Send n wfp_Nil).
        rewrite_wfp_equal. subst. inverts H6...

    - clear H1. inverts H6... inverts H7.
      assert (q0 = wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)).
        rewrite_wfp_equal. subst; clear H4. destruct aq; inverts H8.
      inverts H7... inverts H6... assert (q0 = wfp_Send n wfp_Nil).
        rewrite_wfp_equal. subst. clear H4. inverts H6...
      false. eapply (false_trans_in n). Focus 2. apply H5.
        apply (wfp_cfresh_out _ _ _ _ _ Hfnq H2).     

  + assert (q0 = wfp_Abs a Y (wfp_Par wfp_Nil (wfp_Par (wfp_Abs n X wfp_Nil) (wfp_Send n wfp_Nil)))).
      rewrite_wfp_equal. subst q0. inverts H3.
Qed.

(*********************************
 * PREPARATIONS FOR INPUT NORMAL *
 *********************************)

Lemma wfcgr_tr_in : forall p q, p ≡* q ->
  forall a fp, {{p -- LabIn a ->> AA fp}} ->
    exists fq, {{q -- LabIn a ->> AA fq}} /\
      forall r, (wfp_inst_Abs fp r) ≡* (wfp_inst_Abs fq r).
Proof with automate.
  introv Hcgr. inductions Hcgr. 

  + inductions H; introv Ht; try solve [inverts Ht].

    - lets Heq : wfp_trans_in_chan_eq Ht. subst a0.
      eexists. splits. constructor. introv; simpl.
      assert (wfp_inst_Abs fp r = wfp_subst r X p).
        lets Hyp : wfp_trans_in_abs_eq Ht. inverts Hyp as Hyp...
        destruct Hyp as (Y & HneqY & HfY & Heq). subst fp; simpls.
        rewrite~ <- wfp_subst_decomposition_g.
      rewrite H0. apply wfcgr_subst; constructor~.

    - inverts Ht...
      * destruct ap; inverts H5; rename a0 into fp.
        specializes IHwfcgr_step1 H4. 
        destruct IHwfcgr_step1 as (fp' & Htp' & Hrp').
        eexists; splits. apply act1_in; eassumption.
        introv; simpls. applys* wfcgr_par. constructor~.
      * destruct aq; inverts H5; rename a0 into fq.
        specializes IHwfcgr_step2 H4. 
        destruct IHwfcgr_step2 as (fq' & Htq' & Hrq').
        eexists; splits. apply act2_in; eassumption.
        introv; simpls. applys* wfcgr_par. constructor~.

    - eexists; splits. apply act1_in; eassumption.
      introv; simpls. apply wfcgr_nil_l.

    - inverts Ht... 
      * destruct ap; inverts H3. 
        eexists; splits*. 
        introv; simpls. apply wfcgr_nil_r.
      * inverts H2.

    - inverts Ht...
      * destruct ap; inverts H3.
        eexists; splits. apply act2_in; eassumption.
        introv; simpls. apply wfcgr_par_com.
      * destruct aq; inverts H3.
        eexists; splits. apply act1_in; eassumption.
        introv; simpls. apply wfcgr_par_com.
  
    - inverts Ht...
      * destruct ap; inverts H3.
        eexists; splits. do 2 apply act1_in; eassumption.
        introv; simpls. apply wfcgr_par_assoc_l.
      * destruct aq; inverts H3. inverts H2...
          destruct ap; inverts H4.
          eexists; splits. apply act1_in; apply act2_in; eassumption.
          introv; simpls. apply wfcgr_par_assoc_l.
          destruct aq; inverts H4.
          eexists; splits. apply act2_in; eassumption.
          introv; simpls. apply wfcgr_par_assoc_l.

    - inverts Ht...
      * destruct ap; inverts H3. inverts H2...
          destruct ap; inverts H4.
          eexists; splits. apply act1_in; eassumption.
          introv; simpls. apply wfcgr_par_assoc_r.
          destruct aq; inverts H4.
          eexists; splits. apply act2_in; apply act1_in; eassumption.
          introv; simpls. apply wfcgr_par_assoc_r.
      * destruct aq; inverts H3. 
        eexists; splits. do 2 apply act2_in; eassumption.
        introv; simpls. apply wfcgr_par_assoc_r.

    - eexists; splits*. introv; reflexivity.

  + introv Htx. 
    lets (fy & Hty & Hmy) : IHHcgr1 Htx.
    lets (fz & Htz & Hmz) : IHHcgr2 Hty.
    exists fz; splits*. introv.
    etransitivity; jauto. 
Qed.    

Definition order_oiv (p : wfprocess) :=
  wfp_prod (get_outs p ++ get_ins p ++ get_vars p).

Lemma order_oiv_wfcgr : forall p, p ≡* order_oiv p.
Proof.
  introv. etransitivity. apply ordered_wfcgr. 
  unfold ordered, order_oiv, get_components. 
  etransitivity. apply wfp_prod_wfcgr_app.
  etransitivity. Focus 2. symmetry; apply wfp_prod_wfcgr_app.
  etransitivity. Focus 2. symmetry.
  apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
  etransitivity. Focus 2. apply wfcgr_par_assoc_r.
  etransitivity. Focus 2. apply wfcgr_par_com.
  apply wfcgr_par. reflexivity.
  apply wfp_prod_wfcgr_app.
Qed.

Lemma order_oiv_async_barbed : forall p, async_barbed p (order_oiv p).
Proof.
  introv. apply congr_async_barbed.
  apply order_oiv_wfcgr.
Qed. 

Hint Resolve async_barbed_sym.
Hint Resolve order_oiv_async_barbed.

Lemma async_barbed_to_order_oiv : forall p q,
  async_barbed p q -> async_barbed (order_oiv p) (order_oiv q). 
Proof.
  introv Hyp. 
  transitivity p; auto.
  transitivity q; auto.
Qed.                                    

Lemma order_oiv_cfresh_iff : forall m (p : wfprocess), 
  m ### p <-> m ### (order_oiv p).
Proof.
  introv; splits; intro Hyp; eapply wfcgr_alt_freshCh;
  [ | | symmetry | ]; try apply order_oiv_wfcgr; jauto.
Qed.

(*********** PRELIMINARY ***********

Lemma outs_outs : forall p, get_outs (wfp_prod (get_outs p)) = get_outs p.
Proof.
  introv; lets Hspec : get_outs_spec p. induction (get_outs p).
  + simpls~.
  + lets (p'a & p''a & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite IHl.
      rewrite app_cons. rewrite~ app_nil_l.
      introv Hm; apply Hspec.
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma outs_ins : forall p, get_outs (wfp_prod (get_ins p)) = nil.
Proof.
  introv; lets Hspec : get_ins_spec p. induction (get_ins p).
  + simpls~.
  + lets (aa & Xa & pa & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma outs_vars : forall p, get_outs (wfp_prod (get_vars p)) = nil.
Proof.
  introv; lets Hspec : get_vars_spec p. induction (get_vars p). 
  + simpls; simpl_gets; auto.
  + lets (Xa & Heq & _) : Hspec a. constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. 
      rewrite app_nil_l. apply IHl.
      introv Hm; apply Hspec.
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma ins_outs : forall p, get_ins (wfp_prod (get_outs p)) = nil.
Proof.
  introv; lets Hspec : get_outs_spec p. induction (get_outs p).
  + simpls~.
  + lets (p'a & p''a & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma ins_ins : forall p, get_ins (wfp_prod (get_ins p)) = get_ins p.
Proof.
  introv; lets Hspec : get_ins_spec p. induction (get_ins p). 
  + simpls; simpl_gets; auto.
  + lets (aa & Xa & pa & Heq & _) : Hspec a. constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. 
      rewrite app_cons. rewrite app_nil_l.
      rewrite~ IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma ins_vars : forall p, get_ins (wfp_prod (get_vars p)) = nil.
Proof.
  introv; lets Hspec : get_vars_spec p. induction (get_vars p).
  + simpls~.
  + lets (Xa & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma vars_outs : forall p, get_vars (wfp_prod (get_outs p)) = nil.
Proof.
  introv; lets Hspec : get_outs_spec p. induction (get_outs p).
  + simpls~.
  + lets (p'a & p''a & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma vars_ins : forall p, get_vars (wfp_prod (get_ins p)) = nil.
Proof.
  introv; lets Hspec : get_ins_spec p. induction (get_ins p).
  + simpls~.
  + lets (aa & Xa & pa & Heq & _) : Hspec a. repeat constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

Lemma vars_vars : forall p, get_vars (wfp_prod (get_vars p)) = get_vars p.
Proof.
  introv; lets Hspec : get_vars_spec p. induction (get_vars p). 
  + simpls; simpl_gets; auto.
  + lets (Xa & Heq & _) : Hspec a. constructor. substs. destruct l.
    - simpls; simpl_gets. auto.
    - rewrite wfp_prod_cons. simpl_gets. 
      rewrite app_cons. rewrite app_nil_l.
      rewrite~ IHl. introv Hm; apply Hspec. 
      repeat (inverts Hm as Hm; repeat constructor* ).
Qed.

(*********** INPUTS AND OUTPUTS ***********)

Fixpoint outputs (p : process) :=
  match p with
    | Send a _      => (a :: nil)
    | Par p q       => outputs p ++ outputs q
    | _             => nil
  end.

Lemma outputs_wfp_prod_cons : forall p l1, outputs (wfp_prod (p :: l1)) = outputs p ++ outputs (wfp_prod l1).
Proof.
  introv; destruct l1.  
  + simpls*. rewrite~ app_nil_r.
  + rewrite wfp_prod_cons. simpls*.
Qed.

Lemma outputs_wfp_prod_app : forall l1 l2, outputs (wfp_prod (l1 ++ l2)) = outputs (wfp_prod l1) ++ outputs (wfp_prod l2).
Proof.
  inductions l1; introv.
  
  + rewrite app_nil_l; simpls*.
  + rewrite app_cons. repeat rewrite outputs_wfp_prod_cons. rewrite IHl1.
    rewrite~ app_assoc.
Qed.

Lemma get_ins_no_out : forall p a p' p'', ~ {{wfp_prod (get_ins p) -- LabOut a p'' ->> AP p'}}.
Proof with automate.
  introv. lets Hspec : get_ins_spec p.
  remember (get_ins p) as l; clear Heql. 
  gen a p p' p''. induction l using (measure_induction length); introv Hspec Ht.
  destruct l.

  + simpls; inverts Ht.
  + destruct l.
    - simpls. lets (aw & Xw & pw & Heq & _) : Hspec w. constructor. substs.
      inverts Ht.
    - rewrite wfp_prod_cons in Ht. remember (wfp_prod (w0 :: l)) as l'; inverts Ht...
      * destruct ap; inverts H4.
        lets (aw & Xw & pw & Heq & _) : Hspec w. constructor. substs. inverts H3.
      * destruct aq; inverts H4.
        eapply H. instantiate (1 := (w0 :: l)). rew_length; nat_math.
        introv Hm. apply Hspec. constructor*. eassumption.
Qed.

Lemma get_vars_no_out : forall p a p' p'', ~ {{wfp_prod (get_vars p) -- LabOut a p'' ->> AP p'}}.
Proof with automate.
  introv. lets Hspec : get_vars_spec p.
  remember (get_vars p) as l; clear Heql. 
  gen a p p' p''. induction l using (measure_induction length); introv Hspec Ht.
  destruct l.

  + simpls; inverts Ht.
  + destruct l.
    - simpls. lets (Xw & Heq & _) : Hspec w. constructor. substs.
      inverts Ht.
    - rewrite wfp_prod_cons in Ht. remember (wfp_prod (w0 :: l)) as l'; inverts Ht...
      * destruct ap; inverts H4.
        lets (Xw & Heq & _) : Hspec w. constructor. substs. inverts H3.
      * destruct aq; inverts H4.
        eapply H. instantiate (1 := (w0 :: l)). rew_length; nat_math.
        introv Hm. apply Hspec. constructor*. eassumption.
Qed.

Lemma get_ins_outputs_nil : forall (p : wfprocess), outputs (wfp_prod (get_ins p)) = nil.
Proof.
  introv; lets Hout : get_ins_no_out p; lets Hin : get_ins_spec p.  
  remember (get_ins p) as l. clear Heql. 
  induction l using (measure_induction length). destruct l.

  + simpls~.
   
  + lets (aw & Xw & pw & Heq & _) : Hin w. constructor. substs. destruct l.
    - simpls~. 
    - rewrite wfp_prod_cons in *. remember (w :: l) as l'; clear dependent w.
      simpls. rewrite app_nil_l. apply H. rew_length; nat_math.
      introv Ht. eapply Hout. apply act2_out. eassumption.
      introv Hm. apply Hin. constructor~.                           
Qed.

Lemma get_vars_outputs_nil : forall (p : wfprocess), outputs (wfp_prod (get_vars p)) = nil.
Proof.
  introv; lets Hout : get_vars_no_out p; lets Hin : get_vars_spec p.  
  remember (get_vars p) as l. clear Heql. 
  induction l using (measure_induction length). destruct l.

  + simpls~.
   
  + lets (Xw & Heq & _) : Hin w. constructor. substs. destruct l.
    - simpls~. 
    - rewrite wfp_prod_cons in *. remember (w :: l) as l'; clear dependent w.
      simpls. rewrite app_nil_l. apply H. rew_length; nat_math.
      introv Ht. eapply Hout. apply act2_out. eassumption.
      introv Hm. apply Hin. constructor~.                           
Qed.

Lemma get_outs_wfp_prod_cons : forall l c,
  get_outs (wfp_prod (c :: l)) = get_outs c ++ get_outs (wfp_prod l).
Proof.
  destruct l; introv.
    + simpl; simpl_gets. rewrite~ app_nil_r.
    + rewrite wfp_prod_cons. simpl_gets; auto.
Qed.

Lemma get_ins_wfp_prod_cons : forall l c,
  get_ins (wfp_prod (c :: l)) = get_ins c ++ get_ins (wfp_prod l).
Proof.
  destruct l; introv.
    + simpl; simpl_gets. rewrite~ app_nil_r.
    + rewrite wfp_prod_cons. simpl_gets; auto.
Qed.

Lemma get_vars_wfp_prod_cons : forall l c,
  get_vars (wfp_prod (c :: l)) = get_vars c ++ get_vars (wfp_prod l).
Proof.
  destruct l; introv.
    + simpl; simpl_gets. rewrite~ app_nil_r.
    + rewrite wfp_prod_cons. simpl_gets; auto.
Qed.

Lemma get_ins_outputs : forall a p p' p'', 
  {{p -- LabOut a p'' ->> AP p'}} -> get_ins p = get_ins p'.
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv Ht; try solve [inverts Ht].
  inverts Ht. simpl_gets; auto.
  inverts Ht; subst_wfp.

  destruct ap; inverts H4; simpl_gets. fequals.
  eapply H. solve_wfp_size. eassumption.

  destruct aq; inverts H4; simpl_gets. fequals.
  eapply H. solve_wfp_size. eassumption.
Qed.

Lemma get_vars_outputs : forall a p p' p'', 
  {{p -- LabOut a p'' ->> AP p'}} -> get_vars p = get_vars p'.
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv Ht; try solve [inverts Ht].
  inverts Ht. simpl_gets; auto.
  inverts Ht; subst_wfp.

  destruct ap; inverts H4; simpl_gets. fequals.
  eapply H. solve_wfp_size. eassumption.

  destruct aq; inverts H4; simpl_gets. fequals.
  eapply H. solve_wfp_size. eassumption.
Qed.

(*********** RENAMING ************)

Lemma wf_change_chan : forall a c (p : process), wf (Send a p) -> wf (Send c p).
Proof.
  introv Hwf. inverts Hwf. 
  constructor~.
Qed.

Definition rename_chan_proc (p : wfprocess) (c : chan) :=
match p with
  | mkWfP (Send a p') wfc => mkWfP (Send c p') (wf_change_chan a c p' wfc)
  | _ => p
end.

Lemma rename_chan_proc_eq_send : forall c a p, 
  rename_chan_proc (wfp_Send a p) c = wfp_Send c p.
Proof.
  introv. simpls. rewrite_wfp_equal.
Qed.

Lemma rename_chan_proc_eq_abs : forall c a X p, 
  rename_chan_proc (wfp_Abs a X p) c = wfp_Abs a X p.
Proof.
  introv. simpls~. 
Qed.

Lemma rename_chan_proc_eq_gvar : forall c X, 
  rename_chan_proc (wfp_Gvar X) c = wfp_Gvar X.
Proof.
  introv. simpls~. 
Qed.

Lemma rename_chan_proc_eq_par : forall c p q, 
  rename_chan_proc (wfp_Par p q) c = wfp_Par p q.
Proof.
  introv. simpls~. 
Qed.

Lemma rename_chan_proc_eq_nil : forall c, 
  rename_chan_proc wfp_Nil c = wfp_Nil.
Proof.
  introv. simpls~. 
Qed.

Ltac simpl_rename_chan_proc := repeat
  (repeat rewrite rename_chan_proc_eq_send in *;
   repeat rewrite rename_chan_proc_eq_abs in *;
   repeat rewrite rename_chan_proc_eq_gvar in *;
   repeat rewrite rename_chan_proc_eq_par in *;
   repeat rewrite rename_chan_proc_eq_nil in * ).

Opaque rename_chan_proc.

Fixpoint rename_chan_list (lp : list wfprocess) (lc : list chan) :=
match lp with
 | nil => lp
 | p :: lp' => match lc with 
                | nil => lp
                | c :: lc' => (rename_chan_proc p c) :: rename_chan_list lp' lc'
               end
end.

Lemma rename_chan_list_nil_idem : forall lp, rename_chan_list lp nil = lp.
Proof.
  introv. destruct lp; simpls~.
Qed.

Lemma rename_chan_list_app_split : forall l1 l2 lc,
  exists lc1 lc2, lc = lc1 ++ lc2 /\
  (rename_chan_list (l1 ++ l2) lc = 
   rename_chan_list l1 lc1 ++ rename_chan_list l2 lc2).
Proof.
  inductions l1; introv.

  + repeat (rewrite app_nil_l; simpl). 
    exists (@nil chan) lc. splits*.

  + destruct lc.
    - exists (@nil chan) (@nil chan). splits*.
      rewrite app_cons. simpl. rewrite app_cons.
      fequals. fequals. rewrite~ rename_chan_list_nil_idem.
    - rewrite app_cons. simpl rename_chan_list at 1.
      specializes IHl1 l2 lc. destruct IHl1 as (lc1' & lc2 & Heq1 & Heq2).
      rewrite Heq2. exists (c :: lc1') lc2.
      splits*. rewrite~ app_cons. fequals.
Qed.

Lemma rename_chan_list_no_out : forall lp lc, 
  outputs (wfp_prod lp) = nil -> rename_chan_list lp lc = lp.
Proof.
   induction lp using (measure_induction length); destruct lp; jauto.

   introv Hout; destruct lp; destruct lc; try solve [simpls~].
   simpls; destruct_wf w; simpl_rename_chan_proc; jauto.
   simpls; inverts Hout.

   rewrite wfp_prod_cons in Hout. 
   remember (w0 :: lp) as lp'; clear Heqlp' w0 lp. simpls.
   apply app_eq_nil_inv in Hout. destruct Hout as (Hout_w & Hout_lp').
   destruct_wf w; simpl_rename_chan_proc; [inverts Hout_w | | | |]; 
   fequals; applys~ H; rew_length; nat_math.
Qed.

Lemma rename_chan_list_order_oiv : forall lc p,
  rename_chan_list (get_outs p ++ get_ins p ++ get_vars p) lc = 
  rename_chan_list (get_outs p) lc ++ get_ins p ++ get_vars p.
Proof.
  induction lc using (measure_induction length); destruct lc; introv.

  + repeat rewrite~ rename_chan_list_nil_idem. 
  + lets Hspec_o : get_outs_spec p. 
    lets Hoo : outs_outs p. lets Hio : ins_outs p. 
    lets Hvo : vars_outs p. destruct (get_outs p).
    - simpl; rewrite app_nil_l; clear.
      rewrite~ rename_chan_list_no_out.
      rewrite outputs_wfp_prod_app.
      rewrite get_ins_outputs_nil; rewrite get_vars_outputs_nil.
      rewrite~ app_nil_l.
    - rewrite app_cons. simpl. rewrite app_cons.
      fequals. specializes H lc (wfp_Par (wfp_prod l) 
                                         (wfp_Par (wfp_prod (get_ins p))
                                                  (wfp_prod (get_vars p)))). 
      rew_length; nat_math.
      simpl_gets. rewrite outs_ins, outs_vars, ins_ins, 
                          ins_vars, vars_ins, vars_vars in *.
      replace (get_outs (wfp_prod l)) with l in *.
      replace (get_ins (wfp_prod l)) with (@nil wfprocess) in *.
      replace (get_vars (wfp_prod l)) with (@nil wfprocess) in *.
      repeat rewrite app_nil_l in *; repeat rewrite app_nil_r in *; auto.

      destruct l; jauto.
      rewrite wfp_prod_cons in Hvo; simpl_gets.
      apply app_eq_nil_inv in Hvo. jauto.

      destruct l; jauto.
      rewrite wfp_prod_cons in Hio; simpl_gets.
      apply app_eq_nil_inv in Hio. jauto.

      destruct l; jauto; symmetry.
      rewrite wfp_prod_cons in Hoo; simpl_gets.
      remember (w0 :: l) as l'; clear Heql' w0 l.
      lets (aw & pw & Heqw & _) : Hspec_o w. 
      constructor. subst w.
      simpl_gets. rewrite app_cons in Hoo. rewrite app_nil_l in Hoo.
      inverts Hoo. rewrite~ outs_outs.
Qed.

Lemma rename_chan_list_order_oiv_outs : forall lc p,
   get_outs (wfp_prod (rename_chan_list (get_outs p) lc ++ 
                      get_ins p ++ get_vars p)) = rename_chan_list (get_outs p) lc.
Proof.
   destruct lc; intros.

  + rewrite~ rename_chan_list_nil_idem.
    repeat rewrite get_outs_wfp_prod_app.
    rewrite outs_outs, outs_ins, outs_vars, app_nil_r; auto.

  + lets Hspec : get_outs_spec p. destruct (get_outs p).
    - simpls~. repeat rewrite get_outs_wfp_prod_app. simpls~.
      simpl_gets. rewrite outs_ins, outs_vars.
      repeat rewrite~ app_nil_l.

    - simpl rename_chan_list. rewrite app_cons. 
      lets (aw & pw & Heqw & _) : Hspec w. constructor. subst.
      simpl_rename_chan_proc. rewrite get_outs_wfp_prod_cons.
      simpl_gets. rewrite app_cons, app_nil_l. fequals.
      
      assert (Hspec' : forall q : wfprocess,
          Mem q l -> exists (a : chan) (p' : wfprocess),
            q = wfp_Send a p').
      {{
        introv Hm.
        lets (aq & pq & Heq & _) : Hspec q. repeat constructor~. subst.
        exists~ aq pq.
      }} clear Hspec aw pw c. gen p l.

      destruct lc; intros.
        * repeat rewrite~ rename_chan_list_nil_idem. 
          repeat rewrite get_outs_wfp_prod_app.
          rewrite outs_ins, outs_vars. rewrite app_nil_r.
          clear - Hspec'. induction l. simpls~.
          rewrite get_outs_wfp_prod_cons. 
          lets (aw & pw & Heq) : Hspec' a. constructor. subst. 
          simpl_gets. rewrite app_cons. fequals. 
          rewrite app_nil_l. apply IHl.
          introv Hm. apply Hspec'. repeat constructor~.          

        * destruct l.
            simpl; rewrite app_nil_l.
            rewrite get_outs_wfp_prod_app.
            rewrite outs_ins, outs_vars, app_nil_l. 
            auto.
            
            simpl rename_chan_list. rewrite app_cons.
            rewrite get_outs_wfp_prod_cons.
            lets (aw & pw & Heq) : Hspec' w. constructor. subst.
            simpl_rename_chan_proc. simpl_gets.
            rewrite app_cons, app_nil_l. fequals.
            repeat rewrite get_outs_wfp_prod_app. 
            rewrite outs_ins, outs_vars, app_nil_r.

            assert (Hspec'' : forall q : wfprocess,
                      Mem q l -> exists (a : chan) (p' : wfprocess),
                                   q = wfp_Send a p').
            {{
              introv Hm.
              lets (aq & pq & Heq) : Hspec' q. repeat constructor~. subst.
              exists~ aq pq.
            }} clear Hspec' aw pw c p. gen l.

            induction lc; intros. 
            repeat rewrite~ rename_chan_list_nil_idem. 
            induction l. simpls; simpl_gets; auto.
            lets (aa & pa & Heq) : Hspec'' a. constructor. subst.
            rewrite get_outs_wfp_prod_cons. simpl_gets.
            rewrite app_cons, app_nil_l. fequals. apply IHl.
            introv Hm. apply Hspec''. repeat constructor~.

            destruct l. simpl; simpl_gets; auto.
            simpl rename_chan_list. rewrite get_outs_wfp_prod_cons.
            lets (aw & pw & Heq) : Hspec'' w. constructor. subst.
            simpl_rename_chan_proc. simpl_gets.
            rewrite app_cons, app_nil_l. fequals.
            apply IHlc. introv Hm. apply Hspec''.
            repeat constructor~.
Qed.

Lemma rename_chan_list_order_oiv_ins : forall lc p,
   get_ins (wfp_prod (rename_chan_list (get_outs p) lc ++ 
                      get_ins p ++ get_vars p)) = get_ins p.
Proof.
  intros. 
  lets Hspec : get_outs_spec p. gen lc.
  induction (get_outs p); intros.

  + simpls. rewrite app_nil_l.
    repeat rewrite get_ins_wfp_prod_app.
    rewrite ins_ins, ins_vars, app_nil_r; auto.

  + destruct lc.
    - repeat rewrite~ rename_chan_list_nil_idem. 
      repeat rewrite get_ins_wfp_prod_app.
      rewrite get_ins_wfp_prod_cons. rewrite app_assoc.
      lets (aa & pa & Heq & _) : Hspec a. constructor. subst.
      simpl_gets. rewrite app_nil_l. rewrite ins_ins, ins_vars, app_nil_r.
      replace (get_ins (wfp_prod l)) with (@nil wfprocess).
      rewrite~ app_nil_l.

      clear - Hspec. induction l. simpl; simpl_gets; auto.
      lets (aa' & pa' & Heq & _) : Hspec a. repeat constructor~. subst.
      rewrite get_ins_wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm. apply Hspec. inverts Hm; repeat constructor~. 

    - simpl rename_chan_list. repeat rewrite get_ins_wfp_prod_app.
      rewrite get_ins_wfp_prod_cons. 
      lets (aa & pa & Heq & _) : Hspec a. constructor. subst.
      simpl_rename_chan_proc. simpl_gets. rewrite app_nil_l.
      repeat rewrite <- get_ins_wfp_prod_app. apply IHl.
      introv Hm. apply Hspec. repeat constructor~.
Qed.

Lemma rename_chan_list_order_oiv_vars : forall lc p,
   get_vars (wfp_prod (rename_chan_list (get_outs p) lc ++ 
                      get_ins p ++ get_vars p)) = get_vars p.
Proof.
  intros. 
  lets Hspec : get_outs_spec p. gen lc.
  induction (get_outs p); intros.

  + simpls. rewrite app_nil_l.
    repeat rewrite get_vars_wfp_prod_app.
    rewrite vars_ins, vars_vars, app_nil_l; auto.

  + destruct lc.
    - repeat rewrite~ rename_chan_list_nil_idem. 
      repeat rewrite get_vars_wfp_prod_app.
      rewrite get_vars_wfp_prod_cons. rewrite app_assoc.
      lets (aa & pa & Heq & _) : Hspec a. constructor. subst.
      simpl_gets. rewrite app_nil_l. rewrite vars_ins, vars_vars, app_nil_l.
      replace (get_vars (wfp_prod l)) with (@nil wfprocess).
      rewrite~ app_nil_l.

      clear - Hspec. induction l. simpl; simpl_gets; auto.
      lets (aa' & pa' & Heq & _) : Hspec a. repeat constructor~. subst.
      rewrite get_vars_wfp_prod_cons. simpl_gets. rewrite app_nil_l.
      apply IHl. introv Hm. apply Hspec. inverts Hm; repeat constructor~. 

    - simpl rename_chan_list. repeat rewrite get_vars_wfp_prod_app.
      rewrite get_vars_wfp_prod_cons. 
      lets (aa & pa & Heq & _) : Hspec a. constructor. subst.
      simpl_rename_chan_proc. simpl_gets. rewrite app_nil_l.
      repeat rewrite <- get_vars_wfp_prod_app. apply IHl.
      introv Hm. apply Hspec. repeat constructor~.
Qed.

(*************************************
 * Pointwise async-barbed            *
 *************************************)

(* Pointwise bisimulation of two lists of processes *)
Fixpoint async_barbed_pointwise (lp1 lp2 : list wfprocess) : Prop :=
  match lp1, lp2 with
   | nil, nil             => True
   | p1 :: lp1, p2 :: lp2 => async_barbed p1 p2 /\ async_barbed_pointwise lp1 lp2
   | _, _                 => False
  end.

(* PW bisimulation and length *)
Lemma async_barbed_pw_len : forall lp1 lp2,
  async_barbed_pointwise lp1 lp2 -> length lp1 = length lp2.
Proof.
  inductions lp1; destruct lp2; simpls*.
    introv (_ & H); rew_length; apply IHlp1 in H; nat_math.
Qed.

(* PW bisimulation and wfp_prod *)
Lemma async_barbed_pw_wfp_prod : forall lp1 lp2,
  async_barbed_pointwise lp1 lp2 -> async_barbed (wfp_prod lp1) (wfp_prod lp2).
Proof.
  induction lp1 using (measure_induction length); 
  destruct lp1; destruct lp2; try solve [simpls~]; introv HIOpw; inverts HIOpw.
  destruct lp1; destruct lp2; try solve [simpls~]; try solve [inverts H1].
  repeat rewrite wfp_prod_cons.
  applys~ async_barbed_congr_par.
  applys~ H; rew_length; nat_math.
Qed.

(* 
 *
 * PLAN
 * 
 * 1. Shuffle the outputs of (order_oiv q) so that they match the outputs of p
      (but that is not enough, they need to be async_barbed together)
 *
 * 2. Adapt in_normal_early theorem for this 'version' of q
 *
 *)

Lemma async_barbed_pointwise_output : forall p q,
  async_barbed_pointwise (get_outs p) (get_outs q) ->
    outputs (wfp_prod (get_outs p)) = outputs (wfp_prod (get_outs q)).
Proof.
  intro p. lets Hspec_p : get_outs_spec p. 
  induction (get_outs p).

  + introv Hab. 
    apply async_barbed_pw_len in Hab.
    symmetry in Hab. apply length0 in Hab. 
    rewrite~ Hab. 

  + introv Hab. 
    lets Hspec_q : get_outs_spec q.
    destruct (get_outs q).
    
    - apply async_barbed_pw_len in Hab.
      rew_length in Hab; nat_math.

    - simpl in Hab; repeat rewrite outputs_wfp_prod_cons.
      
      assert (get_outs (wfp_prod l0) = l0).
      {{
        clear - Hspec_q.
        induction l0. simpls; simpl_gets; auto.
        rewrite get_outs_wfp_prod_cons.
        lets (aa & pa & Heq & _) : Hspec_q a. repeat constructor~. subst.
        simpl_gets. rewrite app_cons, app_nil_l. fequals.
        apply IHl0. introv Hm. apply Hspec_q. inverts Hm; repeat constructor~.
      }}
      
      rewrite IHl with (q := wfp_prod l0); try rewrite* H.
      fequals. lets (aw & pw & Heq & _) : Hspec_q w. constructor. subst w.
      lets (aa & pa & Heq' & _) : Hspec_p a. constructor. subst a.
      simpl; fequals. destruct Hab as (Hab & _). clear - Hab.
      tests Hneq : (aa = aw); auto. false.
      apply async_barbed_out_barb_preserving in Hab. specializes Hab aa.
      apply not_not_intro in Hab. apply Hab. rew_logic.
      eexists. eexists. eexists. constructor.
      unfold perform_output. rew_logic. 
      intro x. rew_logic. intro y. intro Hyp. inverts Hyp. false.

      introv Hm. apply Hspec_p. repeat constructor~.
Qed.

Lemma outputs_trans_out_eq : forall (p : wfprocess) a, 
  Mem a (outputs p) <-> exists p' p'', {{p -- LabOut a p'' ->> AP p'}}.
Proof with automate.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv; simpls; splits; introv Hyp; try solve [inverts Hyp];
  try solve [destruct Hyp as (p' & p'' & Ht); inverts Ht; constructor].

  + inverts Hyp. eexists. eexists. constructor. inverts H1.
  + rewrite Mem_app_or_eq in Hyp. inverts Hyp as Hm; 
    rewrite H in Hm; solve_wfp_size; destruct Hm as (p' & p'' & Ht); 
    eexists; eexists.
    - applys* act1_out.
    - applys* act2_out.
  + rewrite Mem_app_or_eq. destruct Hyp as (p' & p'' & Ht). inverts Ht; subst_wfp.
    destruct ap; inverts H4. left; rewrite H...
    destruct aq; inverts H4. right; rewrite H...
Qed.

Lemma async_barbed_no_outs : forall p q, 
  async_barbed p q -> get_outs p = nil -> get_outs q = nil.
Proof.
  introv Hab_pq Houts. apply  async_barbed_to_order_oiv in Hab_pq.
  unfolds order_oiv. rewrite Houts in *. rewrite app_nil_l in *.
  
  replace (get_outs q) with (@nil wfprocess) in *.
  repeat rewrite app_nil_l in *; auto. 
  lets Hspec_q : get_outs_spec q.
  destruct (get_outs q); auto. false.
  symmetry in Hab_pq. lets (aw & pw & Heq & _) : Hspec_q w. 
  constructor. subst. clear Hspec_q.
  apply async_barbed_out_barb_preserving in Hab_pq. specializes Hab_pq aw.
  
  assert (Hyp : perform_output
           (wfp_prod ((wfp_Send aw pw :: l) ++ get_ins q ++ get_vars q)) aw).
  {{
    destruct l; destruct (get_ins q); destruct (get_vars q);
    repeat rewrite app_nil_l; repeat rewrite app_nil_r;
    try rewrite app_cons; repeat rewrite app_nil_l;
    try rewrite wfp_prod_cons;
    try solve [repeat eexists; apply act1_out; constructor].

    simpl. repeat eexists. constructor.
  }}

  specializes Hab_pq Hyp.

  destruct Hyp as (p' & p'' & Ht).
    
  lets Hyp : outputs_trans_out_eq (wfp_prod (get_ins p ++ get_vars p)) aw.
  rewrite outputs_wfp_prod_app in Hyp.
  rewrite get_ins_outputs_nil, get_vars_outputs_nil, app_nil_l in Hyp.  
  destruct Hab_pq as (p1' & p1'' & Ht').
  assert (Mem aw nil) by (rewrite Hyp; repeat eexists; eassumption).
  inverts H.  
Qed.

Lemma removeOne_app_mem_1st : forall (A : Type) l1 l2 (a : A), 
  Mem a l1 -> removeOne a (l1 ++ l2) = (removeOne a l1) ++ l2.
Proof.
  intros A l1. inductions l1; introv Hm; inverts Hm.

  + rewrite app_cons. simpl; cases_if*.
 
  + tests Heq : (a0 = a); rewrite app_cons; simpl; cases_if*.
    rewrite~ IHl1.
Qed.

Lemma removeOne_app_mem_2nd : forall (A : Type) l1 l2 (a : A), 
  Mem a l2 -> permut (removeOne a (l1 ++ l2)) (l1 ++ (removeOne a l2)).
Proof.
  intros A l1. inductions l1; introv Hm.

  + repeat rewrite app_nil_l. apply permut_refl.
 
  + rename a0 into b. tests Heq : (b = a); rewrite app_cons; simpl; cases_if*.
    specializes IHl1 Hm. rewrite app_cons.     
    eapply rem1_permuts. rewrite Mem_app_or_eq. right; eassumption.
    constructor. eapply permut_trans.
    eassumption. simpl; cases_if*.
    rewrite app_cons. specializes IHl1 Hm.
    rewrite~ <- permut_cons_iff.
Qed.

Lemma outputs_trans_out : forall a p p' p'',
  {{p -- LabOut a p'' ->> p'}} -> Mem a (outputs p).
Proof.
  introv Ht; inductions Ht; simpls; try rewrite~ Mem_app_or_eq; constructor.
Qed.

Lemma outputs_trans_out_iter : forall p a p' p'',
  {{p -- LabOut a p'' ->> AP p'}} -> permut (outputs p') (removeOne a (outputs p)).
Proof.
  introv Ht. inductions Ht.
  + simpls. cases_if*.
  + destruct ap; inverts x.
    rename w into p'. specializes IHHt p' (@eq_refl agent).
    simpl. lets Hm : outputs_trans_out Ht.
    rewrite~ removeOne_app_mem_1st.
  + destruct aq; inverts x. rename p'' into q''.
    rename w into q'. specializes IHHt q' (@eq_refl agent).
    simpl. lets Hp : removeOne_app_mem_2nd (outputs p) (outputs q).
    specializes Hp (outputs_trans_out a q _ _ Ht).
    eapply permut_trans. Focus 2. apply permut_sym. eassumption.
    apply~ permut_app_lr.
Qed.

Lemma output_async_barbed_permut : forall p q, async_barbed p q -> permut (outputs p) (outputs q).
Proof with automate.
  introv Hab. lets Hout : async_barbed_out_1 Hab.

  lets Hout_p : outputs_trans_out_eq p.
  lets Hout_q : outputs_trans_out_eq q.

  lets Hiter_p : outputs_trans_out_iter p.
  lets Hiter_q : outputs_trans_out_iter q.  

  remember (outputs p) as op; remember (outputs q) as oq; clear Heqop Heqoq.
  gen oq p q. induction op using (measure_induction length).
  intros; destruct op.

  + clear H; destruct oq; auto.
    symmetry in Hab. apply async_barbed_out_barb_preserving in Hab.
    specializes Hab c. assert (perform_output q c).
    specializes Hout_q c. rewrite <- Hout_q. constructor.
    specializes Hab H. specializes Hout_p c. rewrite <- Hout_p in Hab.
    inverts Hab.

  + assert (Mem c oq).
    {{
      rewrite Hout_q. eapply async_barbed_out_barb_preserving.
      eassumption. rewrite <- (Hout_p c). constructor.
    }}

    lets (oq' & Hoz) : permut_mem_exists H0.
    eapply permut_trans. Focus 2. apply permut_sym; eassumption. 
    rewrite <- permut_cons_iff; rename op into op'.
    assert (perform_output p c).
      specializes Hout_p c; rewrite~ <- Hout_p. constructor.
    destruct H1 as (p' & p'' & Htp).
    lets~ (q' & q'' & Htq & Hab') : Hout Htp.

    apply H with (p := p') (q := q'); jauto.
    - rew_length; nat_math.
    - specializes Hiter_p Htp. simpl in Hiter_p; cases_if*.
      lets Hyp : outputs_trans_out_eq p'.
      introv. specializes Hyp a. rewrite <- Hyp.
      splits; introv Hm.
      * eapply permut_mem. apply permut_sym; eassumption. auto.
      * eapply permut_mem. eassumption. auto.
    - specializes Hiter_p Htp. simpl in Hiter_p; cases_if*.
      lets Hyp : outputs_trans_out_iter p'.
      introv Htp'; specializes Hyp Htp'.
      eapply permut_trans. eassumption.
      apply~ permut_rem1_lr.
    - applys~ async_barbed_out_1.
    - specializes Hiter_q Htq. 
      assert (Hiter_q' : permut (outputs q') oq').
        eapply permut_trans. eassumption.
        eapply permut_trans. eapply permut_rem1_lr.
        eassumption. simpl; cases_if*.
      lets Hyp : outputs_trans_out_eq q'.
      introv. specializes Hyp a. rewrite <- Hyp.
      splits; introv Hm.
      * eapply permut_mem. apply permut_sym; eassumption. auto.
      * eapply permut_mem. eassumption. auto.
    - specializes Hiter_q Htq. 
       assert (Hiter_q' : permut (outputs q') oq').
        eapply permut_trans. eassumption.
        eapply permut_trans. eapply permut_rem1_lr.
        eassumption. simpl; cases_if*.
      lets Hyp : outputs_trans_out_iter q'.
      introv Htq'; specializes Hyp Htq'.
      eapply permut_trans. eassumption.
      apply~ permut_rem1_lr.
Qed.    

Lemma async_barbed_ccinv_cancel_eq : 
  (forall p q C, async_barbed (fill C p) (fill C q) -> async_barbed p q) <-> 
  (forall p q r, async_barbed (wfp_Par p r) (wfp_Par q r) -> async_barbed p q).
Proof.
  splits; introv Hyp. 

  + introv Hab. specializes Hyp p q (CParL CHole r).

  + assert (Hyp' : forall p q r : wfprocess, async_barbed (wfp_Par r p) (wfp_Par r q) → async_barbed p q).
    {{
      introv Hab. eapply Hyp.
      etransitivity. apply async_barbed_congr_par_com. etransitivity. eassumption.
      apply async_barbed_congr_par_com.
    }}

    inductions C; introv Hab; simpls*.
      apply IHC. applys* async_barbed_congr_send_inv.
      apply IHC. applys* async_barbed_congr_abs_inv.
Qed.



(*
Lemma async_barbed_ins_vars_only : forall p q, async_barbed p q ->
  async_barbed (wfp_prod (get_ins p ++ get_vars p)) 
               (wfp_prod (get_ins q ++ get_vars q)).
Proof.
  introv Hab_pq. apply async_barbed_to_order_oiv in Hab_pq.
  unfold order_oiv in *. lets Hspec_p : get_outs_spec p.
  lets Heq_outs : async_barbed_no_outs Hab_pq. gen q.
  induction (get_outs p); intros.

  + rewrite app_nil_l in Hab_pq.
    replace (get_outs q) with (@nil wfprocess) in *.
    repeat rewrite app_nil_l in *; auto. 
    symmetry. repeat rewrite get_outs_wfp_prod_app in Heq_outs.
    repeat rewrite outs_ins, outs_vars in Heq_outs.
    repeat rewrite app_nil_r in *. rewrite outs_outs in Heq_outs.
    jauto.

  + clear Heq_outs. 
    refine (let IHl' := IHl _ in _); clear IHl.
    introv Hm. apply Hspec_p. repeat constructor~. 
    lets (aa & pa & Heq & _) : Hspec_p a. constructor. subst.
    rewrite app_cons in *.

    eapply transitivity in Hab_pq. Focus 2.
    apply congr_async_barbed. symmetry. apply wfp_prod_wfcgr_cons.

    lets Hyp : async_barbed_out_1 Hab_pq aa.
    refine (let Htq := Hyp _ _ _ in _); clear Hyp.
    apply act1_out. constructor. 

    destruct Htq as (q' & q'' & Htq & Hab').
    eapply transitivity in Hab'. Focus 2.
    apply congr_async_barbed. etransitivity.
    Focus 2. apply wfcgr_par_com. apply wfcgr_nil_l.

    lets Hins : get_ins_outputs Htq.
    repeat rewrite get_ins_wfp_prod_app in Hins.
    rewrite ins_outs, ins_ins, ins_vars in Hins.
    rewrite app_nil_l, app_nil_r in Hins.
    repeat rewrite Hins in *. clear Hins.

    lets Hvars : get_vars_outputs Htq.
    repeat rewrite get_vars_wfp_prod_app in Hvars.
    rewrite vars_outs, vars_ins, vars_vars in Hvars.
    repeat rewrite app_nil_l in Hvars.
    repeat rewrite Hvars in *. clear Hvars.

    apply IHl'. etransitivity. eassumption.
    etransitivity. apply congr_async_barbed.
    apply order_oiv_wfcgr. unfolds~ order_oiv. 
    introv Hyp. apply async_barbed_no_outs in Hab'; auto.
    rewrite Hab', app_nil_l.
    repeat rewrite get_outs_wfp_prod_app.
    rewrite outs_ins, outs_vars, app_nil_l. auto.
Qed. *)



Lemma async_barbed_outs_only : forall p q, async_barbed p q ->
  async_barbed (wfp_prod (get_outs p)) (wfp_prod (get_outs q)).
Proof.
  introv Hab_pq. apply async_barbed_to_order_oiv in Hab_pq.
  unfold order_oiv in *. lets Hspec_p : get_outs_spec p.
  lets Heq_outs : async_barbed_no_outs Hab_pq. gen q.
  induction (get_outs p); intros.

  + rewrite app_nil_l in Hab_pq.
    replace (get_outs q) with (@nil wfprocess) in *.
    repeat rewrite app_nil_l in *; auto. 
    symmetry. repeat rewrite get_outs_wfp_prod_app in Heq_outs.
    repeat rewrite outs_ins, outs_vars in Heq_outs.
    repeat rewrite app_nil_r in *. rewrite outs_outs in Heq_outs.
    jauto.
  
  + clear Heq_outs. 
    refine (let IHl' := IHl _ in _); clear IHl.  
    introv Hm. apply Hspec_p. repeat constructor~.
    lets (aa & pa & Heq & _) : Hspec_p a. constructor. subst.
    rewrite app_cons in *.

    eapply transitivity in Hab_pq. Focus 2.
    apply congr_async_barbed. symmetry. apply wfp_prod_wfcgr_cons.

    
    
    lets Hins : get_ins_outputs Htq.
    repeat rewrite get_ins_wfp_prod_app in Hins.
    rewrite ins_outs, ins_ins, ins_vars in Hins.
    rewrite app_nil_l, app_nil_r in Hins.
    repeat rewrite Hins in *. clear Hins.

    
Qed.

Lemma async_barbed_get_outs_permut : forall p q,
  async_barbed p q -> exists goq, permut goq (get_outs q) /\
    async_barbed_pointwise (get_outs p) goq.
Proof.
  
Qed.

zikzik.















(*
 *
 *
 * FINALE
 * 
 * 
 *)

Lemma async_barbed_in_normal_early_order_oiv : 
  forall p q, async_barbed (order_oiv p) (order_oiv q) -> 
    forall a fp, {{(order_oiv p) -- LabIn a ->> AA fp}} ->
       forall m, m ### (order_oiv p) -> m ### (order_oiv q) ->
         exists fq, {{(order_oiv q) -- LabIn a ->> AA fq}}
           /\ async_barbed (wfp_inst_Abs fp (wfp_Send m wfp_Nil))
                           (wfp_inst_Abs fq (wfp_Send m wfp_Nil)).
Proof.
  (* Here we do the three steps on renaming:

       - rename all outputs to fresh ones
       - do the context thingy
       - rename all outputs to original ones
  *)
Qed.

Lemma async_barbed_in_normal_early : in_normal_relation_early async_barbed.
Proof with automate.
  introv Hab Ht Hfmp Hfmq. apply async_barbed_to_order_oiv in Hab.
  lets Hyp : async_barbed_in_normal_early_order_oiv Hab.
  eapply wfcgr_tr_in in Ht; [ | apply order_oiv_wfcgr].
  destruct Ht as (fp' & Htfp' & Hcgr_p').
  rewrite order_oiv_cfresh_iff in Hfmp, Hfmq.
  specializes Hyp Htfp' Hfmp Hfmq.
  destruct Hyp as (fq' & Htfq' & Hab_q').
  eapply wfcgr_tr_in in Htfq'; [ | symmetry; apply order_oiv_wfcgr].
  destruct Htfq' as (fq & Htq & Hcgr_q').

  exists fq; splits~. 
  specializes Hcgr_p' (wfp_Send m wfp_Nil).
  specializes Hcgr_q' (wfp_Send m wfp_Nil).
  etransitivity. apply congr_async_barbed; eassumption.
  etransitivity. Focus 2. apply congr_async_barbed; eassumption.
  auto.
Qed.

Lemma async_barbed_NORbis : forall (p q : wfprocess), 
  closed_process p -> closed_process q -> async_barbed p q -> NORbisC p q.
Proof.
  introv Hcp Hcq Hab. rewrite NORbisC_eNORbisC. splits*. 
    exists async_barbed; repeat splits*.
    apply async_barbed_tau.
    apply async_barbed_out_normal_early. 
    apply async_barbed_in_normal_early.
Qed.

Lemma async_barbed_pipe : forall xs a p q, async_barbed p q -> async_barbed (wfp_pipe a xs p) (wfp_pipe a xs q).
Proof.
  inductions xs; simpls*. 
    introv Hab. specializes IHxs a0 Hab.
    apply async_barbed_context_closed in IHxs.
    specializes IHxs (CAbs a0 a CHole); simpls. apply IHxs.
Qed.

Theorem IObis_async_barbed : forall p q,
  IObis p q <-> async_barbed p q.
Proof.
  introv; splits; intro Hbis.
  
  + applys~ IObis_async_barbed_lr.
  
  + eexists. splits. apply ONORbis_IObis.
    eexists. splits. apply NORbisE_ONORbis.
    introv Hc. apply async_barbed_NORbis; try apply wfp_closed_pipe.
    - introv Hx. rewrite from_list_spec. eapply permut_mem. apply permut_sym; eassumption.
      rewrite <- from_list_spec. rewrite from_to_list. rewrite* in_union.
    - introv Hx. rewrite from_list_spec. eapply permut_mem. apply permut_sym; eassumption.
      rewrite <- from_list_spec. rewrite from_to_list. rewrite* in_union.
    - applys~ async_barbed_pipe.
Qed. 















Fixpoint inputs (p : process) :=
  match p with
    | Receive a _ _ => (a :: nil)
    | Par p q       => inputs p ++ inputs q
    | _             => nil
  end.


Lemma inputs_trans_in : forall a p fp,
  {{p -- LabIn a ->> fp}} -> Mem a (inputs p).
Proof.
  introv Ht; inductions Ht; simpls; try rewrite~ Mem_app_or_eq; constructor.
Qed.



Lemma inputs_trans_in_eq : forall a (p : wfprocess), 
  Mem a (inputs p) <-> exists fp, {{p -- LabIn a ->> AA fp}}.
Proof with automate.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpls; splits; introv Hyp; try solve [inverts Hyp];
  try solve [destruct Hyp as (fp & Ht); inverts Ht; constructor].

  + inverts Hyp. eexists. constructor. inverts H1.
  + rewrite Mem_app_or_eq in Hyp. inverts Hyp as Hm; 
    rewrite H in Hm; solve_wfp_size; destruct Hm as (fp & Ht); eexists.
    - applys* act1_in.
    - applys* act2_in.
  + rewrite Mem_app_or_eq. destruct Hyp as (fp & Ht). inverts Ht; subst_wfp.
    destruct ap; inverts H4. left; rewrite H...
    destruct aq; inverts H4. right; rewrite H... 
Qed.





Lemma inputs_trans_out_eq : forall a p p' p'',
  {{p -- LabOut a p'' ->> AP p'}} -> inputs p = inputs p'.
Proof.
  introv Ht. inductions Ht; simpls*.
  
  destruct ap; inverts x. simpls. rewrite* (IHHt w).
  destruct aq; inverts x. simpls. rewrite* (IHHt w).
Qed.

Lemma tau_channel_match : forall p p',
 {{p -- Tau ->> AP p'}} -> exists a,
 Mem a (inputs p) /\ Mem a (outputs p).
Proof.
  introv Ht.
  lets (a & p1 & p'' & fp & Hout & Hin & _) : wfp_decomposition_tau1 Ht.
  exists a. splits.
    erewrite inputs_trans_out_eq; try eassumption. 
    eapply inputs_trans_in; eassumption.
    eapply outputs_trans_out; eassumption.
Qed.

Lemma no_tau_channels_disjoint : forall (p : wfprocess),
  (from_list (inputs p)) \n (from_list (outputs p)) = \{} <->
  forall p', ~ {{p -- Tau ->> AP p'}}.
Proof with automate.
  introv; splits. 
  + introv Hf Htau. apply tau_channel_match in Htau.
    destruct Htau as (a & Hmi & Hmo); rewrite <- from_list_spec in *.
    assert (a \in \{}). rewrite <- Hf. rewrite* in_inter. 
    rewrite in_empty in H; auto.
  + introv Hnt. apply fset_extens. 
    - introv Hin. rewrite in_inter in Hin.
      destruct Hin as (Hi & Ho); rewrite from_list_spec in *. 
      rewrite outputs_trans_out_eq in Ho.
      destruct Ho as (p' & p'' & Hout). 
      lets Heq : inputs_trans_out_eq Hout. rewrite Heq in Hi; clear Heq.
      rewrite inputs_trans_in_eq in Hi. destruct Hi as (fp' & Hin).
      assert (exists p', {{p -- Tau ->> AP p'}}).
      {{
        clear Hnt; rename x into a. gen a p' p'' fp'. 
        induction p using (measure_induction wfp_size).
        destruct_wf p; intros; try solve [inverts Hout].
        + inverts Hout. inverts Hin.
        + inverts Hout...
          - destruct ap; inverts H4. inverts Hin...
            * destruct ap; inverts H5. specializes H H3 H4; solve_wfp_size.
              destruct H as (p1' & Htau). eexists.
              apply act1_tau. eassumption.
            * destruct aq; inverts H5. eexists. eapply TrTau1; eassumption.
          - destruct aq; inverts H4. inverts Hin...
            * destruct ap; inverts H5. eexists. eapply TrTau2; eassumption.
            * destruct aq; inverts H5. specializes H H3 H4; solve_wfp_size.
              destruct H as (p2' & Htau). eexists.
              apply act2_tau. eassumption.
     }}
     destruct H as (q & Htau); false*.
   - introv Hin. rewrite in_empty in Hin; false.
Qed.

(* Ordered *)

   
(* wfp_prod and congruence *)
Lemma wfp_prod_wfcgr_app:
  forall l1 l2 : list wfprocess,
  wfp_prod (l1 ++ l2) ≡* wfp_Par (wfp_prod l1) (wfp_prod l2).
Proof.
  inductions l1.
    introv; rewrite app_nil_l. simpl. 
    etransitivity. Focus 2. apply wfcgr_par_com. 
    apply wfcgr_nil_l.
    destruct l1. 
      destruct l2. 
        simpls. apply wfcgr_nil_l.
        rewrite app_cons. rewrite app_nil_l.
        rewrite wfp_prod_cons. apply wfcgr_par.
        simpls. reflexivity. reflexivity. 
      introv; repeat rewrite app_cons. 
      repeat rewrite wfp_prod_cons. etransitivity.
      Focus 2. apply wfcgr_par_assoc_l.
      apply wfcgr_par. reflexivity. 
      rewrite <- app_cons. apply IHl1. 
Qed.



Lemma get_ins_get_vars_no_tau : forall p p', ~ {{wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)) -- Tau ->> AP p'}}.
Proof.
  intro p. rewrite <- no_tau_channels_disjoint. simpls.
  apply fset_extens; introv Hin. rewrite in_inter in Hin.
  destruct Hin as (_ & Hin).

  assert (outputs (wfp_prod (get_ins p)) = nil).
  {{
    clear. lets Hspec : get_ins_spec p.
    remember (get_ins p) as l. clear Heql. gen p.
    induction l using (measure_induction length); destruct l; intros; simpls~.
    destruct l.
    + lets (aw & Xw & pw & Heq & _) : Hspec w. constructor. substs~.
    + remember (wfp_prod (w0 :: l)) as l'. simpl. subst.
      rewrite H with (p := p); try (rew_length; nat_math).
      rewrite app_nil_r. lets (aw & Xw & pw & Heq & _) : Hspec w. constructor. substs~.
      introv Hm. apply Hspec. constructor*.
  }}

  assert (outputs (wfp_prod (get_vars p)) = nil).
  {{
    clear. lets Hspec : get_vars_spec p.
    remember (get_vars p) as l. clear Heql. gen p.
    induction l using (measure_induction length); destruct l; intros; simpls~.
    destruct l.
    + lets (Xw & Heq & _) : Hspec w. constructor. substs~.
    + remember (wfp_prod (w0 :: l)) as l'. simpl. subst.
      rewrite H with (p := p); try (rew_length; nat_math).
      rewrite app_nil_r. lets (Xw & Heq & _) : Hspec w. constructor. substs~.
      introv Hm. apply Hspec. constructor*.
  }}

  rewrite H, H0 in *. rewrite from_list_spec in Hin. rewrite app_nil_l in *. inverts Hin.
  rewrite in_empty in Hin. false~.
Qed.





Lemma async_barbed_ccinv_cancel_eq : 
  (forall p q C, async_barbed (fill C p) (fill C q) -> async_barbed p q) <-> 
  (forall p q r, async_barbed (wfp_Par p r) (wfp_Par q r) -> async_barbed p q).
Proof.
  splits; introv Hyp. 

  + introv Hab. specializes Hyp p q (CParL CHole r).

  + assert (Hyp' : forall p q r : wfprocess, async_barbed (wfp_Par r p) (wfp_Par r q) → async_barbed p q).
    {{
      introv Hab. eapply Hyp.
      etransitivity. apply async_barbed_congr_par_com. etransitivity. eassumption.
      apply async_barbed_congr_par_com.
    }}

    inductions C; introv Hab; simpls*.
      apply IHC. applys* async_barbed_congr_send_inv.
      apply IHC. applys* async_barbed_congr_abs_inv.
Qed.

Lemma outputs_get_outs : forall (p : wfprocess) a, Mem a (outputs p) -> exists p', Mem (wfp_Send a p') (get_outs p).
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv Hm; simpl_gets; simpls; try solve [inverts Hm].

  + inverts Hm. eexists. constructor. inverts H1.
  + rewrite Mem_app_or_eq in Hm; inverts Hm as Hm;
    apply H in Hm; solve_wfp_size.
    - destruct Hm as (p' & Hm).
      eexists; rewrite Mem_app_or_eq; left*.
    - destruct Hm as (p' & Hm).
      eexists; rewrite Mem_app_or_eq; right*.
Qed.

Lemma outputs_wfcgr : forall p q, p ≡* q -> permut (outputs p) (outputs q).
Proof with automate.
  introv Hcgr. inductions Hcgr; [ | eapply permut_trans; jauto].

  inductions H; simpls; 
  try solve [try rewrite app_nil_l; try rewrite app_nil_r; repeat rewrite app_assoc; apply permut_refl].
  
  + applys~ permut_app_lr. 
  + apply permut_flip.
Qed.

(* Auxiliary lemmas on gets *)



Lemma ins_trans_out : forall p a p' p'', {{p -- LabOut a p'' ->> AP p'}} ->
  forall q, Mem q (get_ins p') -> Mem q (get_ins p).
Proof with automate.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv Htrans Hm; try solve [inverts Htrans; automate].

  + inverts Htrans...
    - destruct ap; inverts H4; simpl_gets.
      rewrite Mem_app_or_eq in *; inverts Hm; [left | right]; auto.
      eapply H...
    - destruct aq; inverts H4; simpl_gets.
      rewrite Mem_app_or_eq in *; inverts Hm; [left | right]; auto.
      eapply H...
Qed.

Lemma vars_trans_out : forall p a p' p'', {{p -- LabOut a p'' ->> AP p'}} ->
  forall q, Mem q (get_vars p') -> Mem q (get_vars p).
Proof with automate.
  induction p using (measure_induction wfp_size).
  destruct_wf p; introv Htrans Hm; try solve [inverts Htrans; automate].

  + inverts Htrans...
    - destruct ap; inverts H4; simpl_gets.
      rewrite Mem_app_or_eq in *; inverts Hm; [left | right]; auto.
      eapply H...
    - destruct aq; inverts H4; simpl_gets.
      rewrite Mem_app_or_eq in *; inverts Hm; [left | right]; auto.
      eapply H...
Qed.

Lemma get_outs_spec_strong : forall p q, Mem q (get_outs p) ->
  (exists a p', q = wfp_Send a p' /\ exists p'', p ≡* wfp_Par (wfp_Send a p') p'' /\ permut (get_outs p) (q :: get_outs p'')).
Proof with automate.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpl_get_outs; simpls~; introv Hm; try solve [inverts~ Hm].
  inverts Hm. exists a p0. splits~. exists wfp_Nil. splits. constructor*. 
  simpl_gets. apply permut_refl. inverts H1.

  rewrite Mem_app_or_eq in Hm; inverts Hm as Hm; specializes H Hm...
    destruct H as (a & p' & (Heq_q & (p1' & Hcgr_p1 & Hperm))).
    exists a p'. splits~. exists (wfp_Par p1' p2). splits.
    etransitivity. apply wfcgr_par. eassumption. reflexivity.
    constructor. apply WfCgrPar_assoc2.
    simpl_gets. rewrite <- app_cons. applys~ permut_app_lr.

    destruct H as (a & p' & (Heq_q & (p2' & Hcgr_p2 & Hperm))).
    exists a p'. splits~. exists (wfp_Par p1 p2'). splits.
    etransitivity. apply wfcgr_par. reflexivity. eassumption.
    etransitivity. constructor. apply WfCgrPar_assoc1.
    etransitivity. apply wfcgr_par. constructor. apply WfCgrPar_com. reflexivity.
    constructor. apply WfCgrPar_assoc2.    
    simpl_gets. eapply permut_trans.
    apply permut_app_lr. apply permut_refl. eassumption.
    eapply permut_trans. apply permut_flip.
    rewrite app_cons. rewrite <- permut_cons_iff.
    apply permut_flip.
Qed.

Lemma wfcgr_tr_out : forall p q, p ≡* q ->
  forall a, perform_output p a -> perform_output q a.
Proof with automate.
  introv Hc. inductions Hc; jauto.
  inductions H; introv Hout_p; destruct Hout_p as (p1 & p2 & Htp).

  + inverts Htp... repeat eexists. constructor.
  + inverts Htp.
  + inverts Htp...
    - destruct ap; inverts H5.
      lets (p'1 & p'3 & Htp1) : IHwfcgr_step1 a.
      repeat eexists. eassumption.
      repeat eexists. apply act1_out. eassumption.
    - destruct aq; inverts H5.
      lets (p'1 & p'3 & Htp1) : IHwfcgr_step2 a.
      repeat eexists. eassumption.
      repeat eexists. apply act2_out. eassumption.
  + repeat eexists. apply act1_out. eassumption.
  + inverts Htp...
    - destruct ap; inverts H3. repeat eexists. eassumption.
    - inverts H2.
  + inverts Htp...
    - destruct ap; inverts H3.
      repeat eexists. apply act2_out. eassumption.
    - destruct aq; inverts H3.
      repeat eexists. apply act1_out. eassumption.
  + inverts Htp...
    - destruct ap; inverts H3. repeat eexists. repeat applys* act1_out. 
    - destruct aq; inverts H3. inverts H2...
      * destruct ap; inverts H4. repeat eexists. apply act1_out. applys* act2_out.
      * destruct aq; inverts H4. repeat eexists. applys* act2_out. 
  + inverts Htp...
    - destruct ap; inverts H3. inverts H2...
      * destruct ap; inverts H4. repeat eexists. repeat applys* act1_out. 
      * destruct aq; inverts H4. repeat eexists. applys act2_out. applys* act1_out. 
    - destruct aq; inverts H3. repeat eexists. repeat applys* act2_out.    
  + repeat eexists. eassumption.  
Qed.

Lemma wfp_prod_wfcgr_cons : forall p lp, wfp_prod (p :: lp) ≡* wfp_Par p (wfp_prod lp).
Proof.
  destruct lp.
  simpls. apply wfcgr_nil_l.
  rewrite~ wfp_prod_cons. reflexivity.
Qed.

Lemma get_outs_length_async_barbed : forall p q, async_barbed p q -> length (get_outs p) = length (get_outs q).
Proof with automate.
  introv Hab.
  lets Hspec_p : get_outs_spec p.

  assert (Hab_op : async_barbed (ordered p) q).
  {{ 
    etransitivity. symmetry. apply congr_async_barbed. apply ordered_wfcgr. auto.
  }} clear Hab.

  unfolds ordered; unfolds get_components.
  gen q; induction (get_outs p); intros;
  lets Hspec_q : get_outs_spec_strong q.

  assert (Heq_oq : get_outs q = nil).
  {{
    destruct (get_outs q); auto. false.
    lets (aw & pw & Heq & q' & Hcgr & _) : Hspec_q w. constructor. substs.
    clear Hspec_q. 
    symmetry in Hab_op. apply async_barbed_out_barb_preserving in Hab_op.
    specializes Hab_op aw.

    assert (perform_output q aw).
    {{
      eapply wfcgr_tr_out. symmetry; eassumption.
      repeat eexists. applys* act1_out.
    }}

    specializes Hab_op H; clear H.
    rewrite app_nil_l in Hab_op.
    eapply wfcgr_tr_out in Hab_op; try apply wfp_prod_wfcgr_app.
    destruct Hab_op as (p1 & p2 & Ht).
    inverts Ht... 
    + destruct ap; inverts H3. eapply get_vars_no_out; eassumption.
    + destruct aq; inverts H3. eapply get_ins_no_out; eassumption.
  }} rewrite~ Heq_oq.

  lets (aa & pa & Hta & _) : Hspec_p a. constructor. substs. 
  assert (Hab' : async_barbed (wfp_Par (wfp_Send aa pa) (wfp_prod (get_vars p ++ l ++ get_ins p))) q).
  {{
    symmetry. etransitivity. symmetry. eassumption.
    etransitivity. apply congr_async_barbed. apply wfp_prod_wfcgr_app.
    etransitivity. apply congr_async_barbed. apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
    etransitivity. apply congr_async_barbed. apply wfcgr_par. reflexivity. 
    apply wfcgr_par. apply wfp_prod_wfcgr_cons. reflexivity.
    etransitivity. Focus 2. apply congr_async_barbed. apply wfcgr_par. reflexivity. symmetry. apply wfp_prod_wfcgr_app.
    etransitivity. Focus 2. apply congr_async_barbed. apply wfcgr_par. reflexivity. 
    apply wfcgr_par. reflexivity. symmetry. apply wfp_prod_wfcgr_app.
    etransitivity. apply async_barbed_congr_par_assoc_l.
    etransitivity. Focus 2. apply async_barbed_congr_par_assoc_r.
    etransitivity. Focus 2. symmetry. apply async_barbed_congr_par_assoc_l.
    applys~ async_barbed_congr_par.
    etransitivity. apply async_barbed_congr_par_assoc_l.
    applys~ async_barbed_congr_par.
    applys~ async_barbed_congr_par_com.
  }} clear Hab_op.

  lets Hout_q : TrOut aa pa.
  apply act1_out with (p2 := wfp_prod (get_vars p ++ l ++ get_ins p)) in Hout_q.
  eapply async_barbed_out_1 in Hout_q; try eassumption.
  destruct Hout_q as (q' & q'' & Htq & Hab_q).

  assert (Hab_q' : async_barbed (wfp_prod (get_vars p ++ l ++ get_ins p)) q').
  {{
    etransitivity; try eassumption.
    etransitivity. apply async_barbed_congr_nil_l.
    apply async_barbed_congr_par_com.
  }} clear Hab_q.

  specializes IHl Hab_q'.
  introv Hm. eapply Hspec_p. repeat constructor~.

  rew_length. rewrite IHl. gen Htq. clear; intros.
  inductions Htq... simpls; simpl_gets.
  + destruct ap; inverts x.
    specialize (IHHtq w eq_refl).
    simpl_gets; rew_length; nat_math.
  + destruct aq; inverts x.
    specialize (IHHtq w eq_refl).
    simpl_gets; rew_length; nat_math.
Qed.

Lemma inputs_vars : forall p, inputs (wfp_prod (get_vars p)) = nil.
Proof.
  introv. lets Hspec : get_vars_spec p.
  induction (get_vars p). simpls~.
  lets (aa & pa & Heq & _) : Hspec a. constructor. substs~.
  destruct l. simpls~.
  rewrite wfp_prod_cons. remember (w :: l) as l'; clear Heql'.       
  simpls. rewrite app_nil_l. apply IHl. introv Hm.
  apply Hspec. repeat constructor*.
Qed.

Lemma outputs_ins : forall p, outputs (wfp_prod (get_ins p)) = nil.
Proof.
  introv. lets Hspec : get_ins_spec p.
  induction (get_ins p). simpls~.
  lets (aa & Xa & pa & Heq & _) : Hspec a. constructor. substs~.
  destruct l. simpls~.
  rewrite wfp_prod_cons. remember (w :: l) as l'; clear Heql'.       
  simpls. rewrite app_nil_l. apply IHl. introv Hm.
  apply Hspec. repeat constructor*.
Qed.

Lemma outputs_vars : forall p, outputs (wfp_prod (get_vars p)) = nil.
Proof.
  introv. lets Hspec : get_vars_spec p.
  induction (get_vars p). simpls~.
  lets (aa & pa & Heq & _) : Hspec a. constructor. substs~.
  destruct l. simpls~.
  rewrite wfp_prod_cons. remember (w :: l) as l'; clear Heql'.       
  simpls. rewrite app_nil_l. apply IHl. introv Hm.
  apply Hspec. repeat constructor*.
Qed.



Lemma outputs_ordered : forall (p : wfprocess), 
  outputs p = outputs (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))).
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpl_gets; simpls; repeat rewrite~ app_nil_r.

  repeat rewrite outputs_wfp_prod_app;
  repeat rewrite outputs_ins; repeat rewrite outputs_vars.
  repeat rewrite app_nil_r.

  rewrite H; solve_wfp_size.
  rewrite (H p2); solve_wfp_size.
  repeat rewrite outputs_ins; repeat rewrite outputs_vars.
  repeat rewrite~ app_nil_r.
Qed.



(********* n Tau steps *********)

Inductive nTau : wfprocess -> nat -> wfprocess -> Prop :=
 | nTauBase : forall p, nTau p 0 p
 | nTauStep : forall p p', {{p -- Tau ->> AP p'}} -> forall q n, nTau p' n q -> nTau p (S n) q. 

Notation "{{ p -- n -Tau->> q }}" := (nTau p n q) (at level 60).

Hint Resolve nTauBase nTauStep.

Lemma Tau_nTau : forall p p', {{p -- Tau ->> AP p'}} <-> {p -- 1 -Tau->> p'}.
Proof.
  introv; splits*; introv Hyp.
  inverts Hyp. inverts~ H2.
Qed.

Lemma nTau_wfcgr : forall n p p', {{p -- n -Tau->> p'}} -> forall q, p ≡* q -> exists q', {q -- n -Tau->> q'} /\ p' ≡* q'.
Proof with automate.
  induction n using lt_wf_ind; introv Ht Hcgr. destruct n.

  + inverts Ht. exists q; splits*. 

  + inverts Ht. rename p' into p'', p'0 into p'.
    rewrite Tau_nTau in H1. destruct n. inverts H3.
    cut (exists q'', {{q -- Tau ->> AP q''}} /\ p'' ≡* q'').
    - introv (q'' & Htauq & Hcgrq). exists q''; splits*.
    - rewrite <- Tau_nTau in H1. eapply congr_tau in H1; try eassumption. 
    - lets (q' & Ht1 & Hcgr1) : H H1 Hcgr. nat_math.
      lets (q'' & Ht2 & Hcgr2) : H H3 Hcgr1. nat_math.
      eexists. splits; try eassumption. econstructor; try eassumption.
      rewrite~ Tau_nTau.
Qed.

Lemma async_barbed_nTau : forall n p p', {{p -- n -Tau->> p'}} -> forall q, async_barbed p q -> 
                            exists q', {{q -- n -Tau->> q'}} /\ async_barbed p' q'.
Proof.
  induction n using (lt_wf_ind); introv Htau_p Hab; destruct n.

  + exists q; splits*. inverts~ Htau_p.
  + inverts Htau_p. 
    lets (q' & Htau_q & Hab_q) : async_barbed_tau Hab H1.
    specializes H H3 Hab_q. 
    destruct H as (q'' & Htau_q'' & Hab_q'').
    exists q''; splits*.
Qed.

Fixpoint wfp_prod_abs_chan lco lcr X :=
match lco, lcr with
 | (a :: lco), (b :: lcr) => wfp_Par (wfp_Abs a X (wfp_Send b (wfp_Gvar X))) (wfp_prod_abs_chan lco lcr X)
 | _, _ => wfp_Nil
end.

Lemma wfp_prod_abs_chan_nil_l : forall lcr X, wfp_prod_abs_chan nil lcr X = wfp_Nil.
Proof.
  introv. destruct lcr; simpls~.
Qed.

Lemma wfp_prod_abs_chan_nil_r : forall lco X, wfp_prod_abs_chan lco nil X = wfp_Nil.
Proof.
  introv. destruct lco; simpls~.
Qed.

Lemma outputs_wfp_prod_abs_chan : forall lco lcr X, outputs (wfp_prod_abs_chan lco lcr X) = nil.
Proof.
  inductions lco; introv.

  + rewrite wfp_prod_abs_chan_nil_l. simpls~.
  + destruct lcr. rewrite wfp_prod_abs_chan_nil_r. simpls~.
    simpls. rewrite app_nil_l. apply IHlco.
Qed.


(* 
   Initial renaming lemma: 

   If we take a process and a list of n fresh names, and wrap the process
   in a context as described in the paper, then after n tau steps, we can arrive
   at a process where the first n outputs have been renamed with the fresh names

   n <= actual number of outputs in the process
*)

Lemma renaming_nTau : forall X lc, noDup lc -> forall (p : wfprocess), length lc <= length (outputs p) ->
  (Forall (fun c => c ### p) lc) -> 
  exists p', {{(fill (CParL CHole (wfp_prod_abs_chan (outputs p) lc X)) p) -- (length lc) -Tau->> p'}} /\ 
             p' ≡* wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))).
Proof with automate.
   inductions lc; introv Hnd Hlc Hflc; simpls.

  + rewrite wfp_prod_abs_chan_nil_r. 
    rew_length. eexists. splits. constructor.
    etransitivity. apply wfcgr_nil_r.
    etransitivity. apply ordered_wfcgr.
    unfolds ordered; unfolds get_components.
    etransitivity. apply wfp_prod_wfcgr_app.
    etransitivity. apply wfcgr_par_com.
    etransitivity. Focus 2. apply wfcgr_par_assoc_r.
    apply wfcgr_par; try reflexivity.
    rewrite <- rename_chan_list_nil_idem. 
    apply wfp_prod_wfcgr_app.

  + cut (exists p', {{wfp_Par (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) (wfp_prod_abs_chan (outputs p) (a :: lc) X) -- length (a :: lc) -Tau->> p'}} /\ p' ≡* wfp_Par (wfp_prod (rename_chan_list (get_outs p) (a :: lc))) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))).
    {{
      introv (p' & Ht & Hcgr).
      lets (p'' & Htp'' & Hcgr'') : nTau_wfcgr Ht (wfp_Par p (wfp_prod_abs_chan (outputs p) (a :: lc) X)).
      {{
        apply wfcgr_par; try reflexivity.
        etransitivity. Focus 2. symmetry. apply ordered_wfcgr.
        unfolds ordered; unfolds get_components.
        etransitivity. apply wfcgr_par_assoc_l.
        etransitivity. apply wfcgr_par_com.
        etransitivity. Focus 2. symmetry. apply wfp_prod_wfcgr_app.
        apply wfcgr_par. reflexivity. symmetry. apply wfp_prod_wfcgr_app.
      }}
    
      eexists. splits. eassumption. etransitivity; try eassumption.
      symmetry; auto.
    }}

    destruct Hnd as (Hnm & Hnd). specializes IHlc Hnd. 
    remember (outputs p) as OP.
    rewrite outputs_ordered in *.

    simpl outputs in HeqOP. rewrite outputs_ins, outputs_vars in *.
    repeat rewrite app_nil_r in *.

    symmetry in HeqOP; destruct OP.
    - assert (length (a :: lc) = 0) by (rew_length in *; nat_math). rewrite H. 
      eexists. splits. constructor.
      rewrite wfp_prod_abs_chan_nil_l.
      etransitivity. apply wfcgr_nil_r. 
      apply wfcgr_par; try reflexivity.
      lets Hspec_p : get_outs_spec p.
      destruct (get_outs p).
      simpls. reflexivity. false. 

    - replace (length (a :: lc)) with (S (length lc)) by (rew_length; nat_math). simpls.
      lets Hspec_p : get_outs_spec p.

      assert (Hf_out : Forall (fun c => c ### (wfp_prod (get_outs p))) lc).
      {{
        inverts Hflc. rewrite <- Forall_Mem in *. 
        introv Hm; specializes H2 Hm. 
        eapply wfcgr_alt_freshCh in H2. Focus 2. apply ordered_wfcgr.
        unfolds ordered; unfolds get_components.
        eapply wfcgr_alt_freshCh in H2. Focus 2. apply wfp_prod_wfcgr_app.
        eapply wfcgr_alt_freshCh in H2. Focus 2. 
        apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
        simpls. repeat rewrite cfresh_par_iff in *. jauto.
      }}

      destruct (get_outs p). simpls. inverts HeqOP.
      lets (aw & pw & Heq & _) : Hspec_p w. constructor. substs.
      rewrite outputs_wfp_prod_cons in HeqOP. simpl in HeqOP.
      rewrite app_cons in HeqOP; rewrite app_nil_l in HeqOP. inverts HeqOP. 
      simpl rename_chan_list. simpl_rename_chan_proc.
      
      destruct l. simpls. eexists. splits. econstructor. eapply TrTau1.
      apply act1_out. constructor. apply act1_in. constructor.
      assert (length lc = 0) by (rew_length in *; nat_math). rewrite H.
      constructor. simpls... cases_if*. clear e.
      etransitivity. etransitivity. apply wfcgr_par. etransitivity.
      apply wfcgr_par_com. apply wfcgr_nil_r. apply wfcgr_nil_r. apply wfcgr_par_com.
      reflexivity.

      repeat rewrite wfp_prod_cons. remember (w :: l) as l'. clear Heql' w l.
       
      lets Hout : TrOut c pw.
      apply act1_out with (p2 := wfp_prod l') in Hout.
      apply act1_out with (p2 := (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) in Hout.
      lets Hin : TrIn c X (wfp_Send a (wfp_Gvar X)).
      apply act1_in with (p2 := wfp_prod_abs_chan (outputs (wfp_prod l')) lc X) in Hin.
      lets Htau : TrTau1 Hout Hin. simpl in Htau... cases_if*; clear e Hin Hout.

      assert (wfp_prod_abs_chan (outputs (wfp_prod l')) lc X = 
              wfp_prod_abs_chan (outputs (wfp_prod l') & a) lc X).
      {{
        rew_length in *. rewrite le_SS in Hlc. 
        gen Hlc. clear. remember (wfp_prod l') as p. clear.

        gen lc a X. induction (outputs p); introv Hl. 
        assert (length lc = 0) by (rew_length in *; nat_math).
        apply length0 in H; subst. repeat rewrite~ wfp_prod_abs_chan_nil_r.

        rewrite app_cons. destruct l.
        destruct lc. repeat rewrite~ wfp_prod_abs_chan_nil_r.
        assert (lc = nil). apply length0. rew_length in *; nat_math.
        substs. simpls~. rewrite app_cons. 
        destruct lc. repeat rewrite~ wfp_prod_abs_chan_nil_r.
        remember (c0 :: l) as l1. remember (c0 :: l & a) as l2. simpl.
        subst l1 l2. fequals. apply IHl. rew_length in *. nat_math.
      }}
      
      (* GET THE P RIGHT! *)

      remember (wfp_Par (wfp_Par (wfp_Par wfp_Nil (wfp_prod l'))
                                 (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                        (wfp_Send a pw)) as P.

      assert (outputs P = outputs (wfp_prod l') & a).
      {{
        subst P. simpl. rewrite outputs_ins, outputs_vars.
        repeat rewrite app_nil_l, app_nil_r. auto.
      }}
      
      lets (P' & HtP' & HcgrP') : IHlc P.
        rewrite H0. rew_length in *. nat_math.
        inverts Hflc. rewrite <- Forall_Mem in *. 
        introv Hm; specializes H4 Hm. lets H5 : H4.
        eapply wfcgr_alt_freshCh in H4. Focus 2. apply ordered_wfcgr.
        unfolds ordered; unfolds get_components.
        eapply wfcgr_alt_freshCh in H4. Focus 2. apply wfp_prod_wfcgr_app.
        eapply wfcgr_alt_freshCh in H4. Focus 2. 
        apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.

        rewrite HeqP. simpls. 
        repeat rewrite cfresh_par_iff in *; repeat splits; fresh_solve; jauto.
        
        inverts H4. inverts H2.

        specializes Hf_out Hm. gen Hf_out. clear.
        destruct l'; introv Hf; simpls; fresh_solve.       

        rewrite cfresh_send. splits.
          lets (ar & pr & Heq & (p' & Hcgr)) : Hspec_p (wfp_Send c pw).
          constructor. inverts Heq... eapply wfcgr_alt_freshCh in H5.
          Focus 2. eassumption. simpls. rewrite cfresh_par_iff in H5.
          inverts H5; fresh_solve.
          introv Heq; substs. false*.

        assert (Hcgr : (wfp_Par P (wfp_prod_abs_chan (outputs P) lc X)) ≡* 
                       wfp_Par (wfp_Par (wfp_Par wfp_Nil (wfp_prod l'))
                                        (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                               (wfp_Par (wfp_Send a pw)
                                        (wfp_prod_abs_chan (outputs (wfp_prod l')) lc X))).
        {{
          rewrite HeqP at 1. rewrite H. rewrite <- H0.
          apply wfcgr_par_assoc_r. 
        }}

        lets HtP : nTau_wfcgr HtP' Hcgr.
        destruct HtP as (Q & HtQ & HcgrQ).

        assert (Houts_P : get_outs P = l' & (wfp_Send a pw)).
        {{
          rewrite HeqP. simpl_gets. rewrite outs_ins, outs_vars.
          fequals. rewrite app_nil_l. repeat rewrite app_nil_r.
          gen Hspec_p. clear. inductions l'; intros. simpls*.
          lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
          destruct l'. simpls*.
          rewrite wfp_prod_cons. remember (w :: l') as l''; clear Heql'' w l'.
          simpl_gets. rewrite app_cons. rewrite app_nil_l. fequals.
          apply IHl'. introv Hm. apply Hspec_p. inverts Hm; repeat constructor~.
        }}

        assert (Hins_P : get_ins P = get_ins p).
        {{
          rewrite HeqP. simpl_gets. rewrite ins_ins, ins_vars.
          repeat rewrite app_nil_l; repeat rewrite app_nil_r.
          assert (get_ins (wfp_prod l') = nil).
          {{
            gen Hspec_p. clear. inductions l'; intros. simpls*.
            lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
            destruct l'. simpls*.
            rewrite wfp_prod_cons. remember (w :: l') as l''; clear Heql'' w l'.
            simpl_gets. rewrite app_nil_l. apply IHl'.
            introv Hm. apply Hspec_p. inverts Hm; repeat constructor~.
          }}
          rewrite H1. rewrite~ app_nil_l.
        }}

        assert (Hvars_P : get_vars P = get_vars p).
        {{
          rewrite HeqP. simpl_gets. rewrite vars_ins, vars_vars.
          repeat rewrite app_nil_l; repeat rewrite app_nil_r.
          assert (get_vars (wfp_prod l') = nil).
          {{
            gen Hspec_p. clear. inductions l'; intros. simpls*.
            lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
            destruct l'. simpls*.
            rewrite wfp_prod_cons. remember (w :: l') as l''; clear Heql'' w l'.
            simpl_gets. rewrite app_nil_l. apply IHl'.
            introv Hm. apply Hspec_p. inverts Hm; repeat constructor~.
          }}
          rewrite H1. rewrite~ app_nil_l.
        }}

        eexists. splits. econstructor; eassumption.         
        etransitivity. symmetry; eassumption. 
        etransitivity. apply HcgrP'. apply wfcgr_par.
        
        rewrite Houts_P. assert (length lc <= length l').
        {{
          rew_length in *. rewrite le_SS in Hlc.
          etransitivity. eassumption. gen Hspec_p. clear.
          inductions l'; intros. simpls; rew_length; nat_math.
          lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
          destruct l'. simpls; rew_length; nat_math.
          rewrite wfp_prod_cons. remember (w :: l') as l''; clear Heql'' w l'.
          simpls. rew_length. rewrite le_SS. apply IHl'.
          introv Hm; apply Hspec_p. inverts Hm; repeat constructor~.
        }}

        gen H1 Hspec_p. clear; intros.
        destruct lc. repeat rewrite <- rename_chan_list_nil_idem.
        etransitivity. apply wfp_prod_wfcgr_app. etransitivity.
        Focus 2. symmetry. apply wfp_prod_wfcgr_cons. apply wfcgr_par_com.
        etransitivity. Focus 2. symmetry. apply wfp_prod_wfcgr_cons.
        etransitivity. Focus 2. apply wfcgr_par_com.

        remember (c0 :: lc) as l. clear Heql c0 lc.
        gen l a p pw c. inductions l'; intros.

        assert (l = nil). apply length0. rew_length in *; nat_math.
        subst. simpls. etransitivity. apply wfcgr_nil_l.
        apply wfcgr_par_com.

        rewrite app_cons. destruct l. 
        repeat rewrite <- rename_chan_list_nil_idem. 
        etransitivity. etransitivity. apply wfp_prod_wfcgr_cons.
        apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
        etransitivity. Focus 2. apply wfcgr_par.
        symmetry. apply wfp_prod_wfcgr_cons. reflexivity.
        simpls. etransitivity. apply wfcgr_par_assoc_l.
        apply wfcgr_par; reflexivity.
        
        lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.     
        simpl rename_chan_list. simpl_rename_chan_proc. 
        etransitivity. apply wfp_prod_wfcgr_cons.
        etransitivity. Focus 2. symmetry. apply wfcgr_par.
        apply wfp_prod_wfcgr_cons. reflexivity.
        etransitivity. Focus 2. apply wfcgr_par_assoc_l.
        apply wfcgr_par. reflexivity. eapply IHl'.
        rew_length in *. nat_math.
        introv Hm. apply Hspec_p. inverts Hm; repeat constructor~.
        
        rewrite Hins_P, Hvars_P. reflexivity.
Qed.

Lemma inputs_outs : forall p, inputs (wfp_prod (get_outs p)) = nil.
Proof.
  introv. lets Hspec_p : get_outs_spec p. induction (get_outs p).
  simpls~. lets (aa & Xa & pa & Heq & _) : Hspec_p a. constructor. substs. 
  destruct l. simpls~. rewrite wfp_prod_cons. 
  remember (w :: l) as l'; clear Heql' w l.
  simpls. rewrite app_nil_l. apply IHl. introv Hm.
  apply Hspec_p. repeat constructor~.
Qed.

Lemma inputs_rename_outs : forall lc p, 
  inputs (wfp_prod (rename_chan_list (get_outs p) lc)) = nil.
Proof.
  assert (Hspec : forall p lc q, Mem q (rename_chan_list (get_outs p) lc) ->
            exists aq pq, q = wfp_Send aq pq).
  {{
    intro p. lets Hspec : get_outs_spec p.
    induction (get_outs p); introv Hm. simpls. inverts Hm.

    assert ((∀ q : wfprocess,
         Mem q l
         → ∃ (a : chan) (p' : wfprocess),
           q = wfp_Send a p'
           ∧ (∃ p'' : wfprocess, p ≡* wfp_Par (wfp_Send a p') p''))).
    {{
      introv Hm'. apply Hspec. repeat constructor~.
    }} specializes IHl H. clear H.

    lets (aq & pq & Heq & _) : Hspec a. constructor. substs.
    destruct lc. rewrite <- rename_chan_list_nil_idem in *. 
    inverts Hm as Hm. repeat eexists. 
    lets (aq' & pq' & Heq & _) : Hspec q. repeat constructor~. substs.
    repeat eexists.


    simpls. simpl_rename_chan_proc. inverts Hm as Hm. repeat eexists.
    eapply IHl. eassumption.
  }}

  introv. specializes Hspec p lc. 
  induction (rename_chan_list (get_outs p) lc). simpls~.
  lets (aa & pa & Heq) : Hspec a. constructor. subst.
  destruct l. simpls~. rewrite wfp_prod_cons.
  remember (w :: l) as l'; clear Heql' w l.
  simpls. rewrite app_nil_l. apply IHl. introv Hm.
  apply Hspec. repeat constructor~.
Qed.

Lemma inputs_ins : forall p, inputs (wfp_prod (get_ins p)) = inputs p.
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpls~.

  simpl_gets. rewrite <- (H p1); solve_wfp_size.
  rewrite <- (H p2); solve_wfp_size. clear H.

  lets H : (get_ins_spec p1). induction (get_ins p1).
  + simpl. repeat rewrite~ app_nil_l.
  + lets (aa & Xa & pa & Heq & _) : H a. constructor. substs. destruct l. 
    - rewrite app_cons. rewrite app_nil_l.
      destruct (get_ins p2). simpls. rewrite~ app_nil_r.
      rewrite wfp_prod_cons. simpls~.
    - repeat rewrite app_cons. repeat rewrite wfp_prod_cons. 
      rewrite <- app_cons. remember (w :: l) as l'. clear Heql' w l.
      simpls. repeat rewrite app_cons. repeat rewrite app_nil_l.
      fequals. apply IHl. introv Hm; apply H. repeat constructor~.              
Qed.

Lemma outputs_get_outs_length : forall (p : wfprocess), (length (outputs p) = length (get_outs p)).
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpls; simpl_gets; rew_length in *; try nat_math.
    rewrite H; solve_wfp_size. rewrite H; solve_wfp_size.
Qed.

Lemma renaming_noTau : forall lc (p : wfprocess), 
  length lc >= length (outputs p) ->
    (Forall (fun c => c ### p) lc) -> forall p',
      ~{{wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) -- Tau ->> AP p'}}.
Proof.
  introv Hl Hf. rewrite <- no_tau_channels_disjoint.
  symmetry; apply fset_extens. introv Hin. rewrite in_empty in Hin; false.
  simpls. rewrite inputs_vars, outputs_ins, outputs_vars. 
  repeat rewrite app_nil_r. rewrite inputs_rename_outs, inputs_ins.
  rewrite app_nil_l. introv Hm. rewrite in_inter in Hm; 
  repeat rewrite from_list_spec in *. destruct Hm as (Hmi & Hmo). false.

  rewrite outputs_get_outs_length in Hl. 

  assert (Hfx : x ### p). 
  {{
    clear Hmi. gen lc x. lets Hspec : get_outs_spec p.
    induction (get_outs p); intros. simpls. inverts Hmo.

    lets (aa & pa & Heq & _) : Hspec a. constructor. substs.
    destruct lc. false. rew_length in *. nat_math.
    simpls rename_chan_list. simpl_rename_chan_proc.
    rewrite outputs_wfp_prod_cons in Hmo. rewrite Mem_app_or_eq in Hmo.
    inverts Hmo. simpls. inverts H. inverts~ Hf. inverts H1.
    apply IHl with (lc := lc). introv Hm. apply Hspec. repeat constructor~.
    rew_length in *; nat_math. inverts~ Hf. auto.
  }}

  gen Hmi Hfx. clear. induction p using (measure_induction wfp_size).
  destruct_wf p; intros; try solve [inverts Hmi].
  
  inverts Hmi. rewrite wfp_cfresh_Abs in Hfx. jauto. inverts H1.
  simpls. rewrite cfresh_par_iff in Hfx. rewrite Mem_app_or_eq in Hmi.
  inverts Hmi; eapply H; try eassumption; solve_wfp_size; jauto.
Qed.

Lemma outputs_get_outs_eq : forall (p : wfprocess), outputs p = outputs (wfp_prod (get_outs p)).
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpl_gets; simpls~.

  rewrite (H p1); solve_wfp_size.
  rewrite (H p2); solve_wfp_size.
  rewrite~ outputs_wfp_prod_app.
Qed.

Lemma length_split : forall (A : Type) (l1 l2 l : list A), 
  length l = length (l1 ++ l2) ->
    exists l' l'', length l' = length l1 /\ length l'' = length l2 /\ l = l' ++ l''.
Proof.
  intro A. inductions l1; introv Hl.
  
  + exists (@nil A) l; splits*.

  + rewrite app_cons in Hl. destruct l.
    - symmetry in Hl. apply length0 in Hl. inverts Hl.
    - lets (l' & l'' & Hl1 & Hl2 & Heq) : IHl1 l2 l.
      rew_length in *. nat_math.
      exists (a0 :: l') l''; splits*.
      rew_length in *; nat_math.
      rewrite app_cons. rewrite~ Heq.
Qed.

Lemma wfp_prod_abs_chan_app : forall l1 l2 lc1 lc2 X, 
  length l1 = length lc1 -> length l2 = length lc2 ->
  wfp_prod_abs_chan (l1 ++ l2) (lc1 ++ lc2) X ≡* 
  wfp_Par (wfp_prod_abs_chan l1 lc1 X) (wfp_prod_abs_chan l2 lc2 X).
Proof.
  inductions l1; introv Hl1 Hl2.

  + symmetry in Hl1. apply length0 in Hl1. substs.
    repeat rewrite app_nil_l. rewrite wfp_prod_abs_chan_nil_l.
    etransitivity. apply wfcgr_nil_l. apply wfcgr_par_com.

  + destruct lc1. apply length0 in Hl1. inverts Hl1.
    repeat rewrite app_cons. simpls. etransitivity.
    Focus 2. apply wfcgr_par_assoc_l. apply wfcgr_par. reflexivity.
    applys~ IHl1.
Qed.

Lemma wfp_prod_abs_chan_permut_one : forall lp lc X lp', length lc = length lp ->
  permut_one lp lp' -> exists lc', permut_one lc lc' /\ 
  wfp_prod_abs_chan lp lc X ≡* wfp_prod_abs_chan lp' lc' X.
Proof.
  introv Hl Hp. inverts Hp.
  lets (l1c & l' & Hl1c & Hl' & Heq) : length_split Hl. substs.
  lets (l2c & l'' & Hl2c & Hl'' & Heq) : length_split Hl'. substs.
  lets (l3c & l4c & Hl3c & Hl4c & Heq) : length_split Hl''. substs.

  eexists. splits. constructor. clear Hl Hl' Hl''.
  etransitivity; try applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  etransitivity. Focus 2. symmetry. try applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  apply wfcgr_par. reflexivity.

  etransitivity; try applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  etransitivity. Focus 2. symmetry. try applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  etransitivity. apply wfcgr_par. reflexivity.
  applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  etransitivity. Focus 2. symmetry. apply wfcgr_par. reflexivity.
  applys* wfp_prod_abs_chan_app; try (rew_length; nat_math).
  etransitivity. apply wfcgr_par_com. etransitivity. apply wfcgr_par_assoc_r.
  apply wfcgr_par. reflexivity. apply wfcgr_par_com.
Qed.
    
Lemma wfp_prod_abs_chan_permut_rtclosure : forall lp lc X lp', length lc = length lp -> 
  rtclosure (permut_one (A:=chan)) lp lp' -> exists lc', rtclosure (permut_one (A:=chan)) lc lc' /\ 
  wfp_prod_abs_chan lp lc X ≡* wfp_prod_abs_chan lp' lc' X.
Proof.
  introv Hl Hc. gen lc X. inductions Hc; introv Hl; introv.
  exists lc; splits. constructor. reflexivity.
  assert (length x = length y).
    apply permut_len. eapply rtclosure_step. eassumption. constructor.

  apply wfp_prod_abs_chan_permut_one with (lc := lc) (X := X) in H; auto.
  destruct H as (lc' & Hp' & Hcgr'). clear Hc.

  assert (length lc = length lc').
    apply permut_len. eapply rtclosure_step. eassumption. constructor.

  specializes IHHc lc' X. nat_math. destruct IHHc as (lc'' & Hc'' & Hcgr'').
  exists lc''. splits*. eapply rtclosure_step; eassumption. 
  etransitivity; eassumption.
Qed.

Lemma wfp_prod_abs_chan_permut : forall lp lc X lp', length lc = length lp ->
  permut lp lp' -> exists lc', permut lc lc' /\ 
  wfp_prod_abs_chan lp lc X ≡* wfp_prod_abs_chan lp' lc' X.
Proof.
  introv Hl Hp. gen lc X. inverts Hp; introv Hl; introv.

  + exists lc; splits*. reflexivity.

  + lets H' : H; apply wfp_prod_abs_chan_permut_one with (lc := lc) (X := X) in H; auto.
    destruct H as (lc' & Hp' & Hcgr'). 
    apply wfp_prod_abs_chan_permut_rtclosure with (lc := lc') (X := X) in H0.
    destruct H0 as (lc'' & Hp'' & Hcgr'').
    exists lc''. splits*. eapply rtclosure_step; eassumption.
    etransitivity; eassumption.
    transitivity (length lp). Focus 2. apply permut_len. 
    eapply rtclosure_step. eassumption. constructor.
    transitivity (length lc); auto. symmetry.
    apply permut_len. eapply rtclosure_step. eassumption. constructor.
Qed.

Lemma wfp_prod_abs_chan_permut_outputs : forall lc X p q, async_barbed p q -> 
  length lc = length (outputs p) -> exists lc', permut lc lc' /\
  wfp_prod_abs_chan (outputs p) lc X ≡* wfp_prod_abs_chan (outputs q) lc' X.
Proof.
  introv Hab. apply output_async_barbed in Hab.
  remember (outputs p) as lp; remember (outputs q) as lq.
  clear dependent p. clear dependent q.
  introv Hl; applys~ wfp_prod_abs_chan_permut.
Qed.

Lemma outputs_outs : forall p, outputs (wfp_prod (get_outs p)) = outputs p.
Proof.
  induction p using (measure_induction wfp_size).
  destruct_wf p; simpls~.

  simpl_gets. rewrite outputs_wfp_prod_app. rewrite <- (H p1); solve_wfp_size.
  rewrite~ <- (H p2); solve_wfp_size. 
Qed.

(*
  Processes that are async_barbed follow renaming
*)
Lemma renaming_part_1 : forall p q, async_barbed p q -> 
  forall lc, noDup lc -> 
             length lc = length (get_outs p) ->
             (Forall (fun c => c ### p) lc)   -> 
             (Forall (fun c => c ### q) lc)   ->
    exists lc', permut lc lc' /\
    async_barbed (wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                 (wfp_Par (wfp_prod (rename_chan_list (get_outs q) lc')) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))).
Proof.
  introv Hab_p_q Hnd Hllcp Hflcp Hflcq.

  assert (Hllcq : length lc = length (get_outs q)).
  {{
    rewrite <- outputs_get_outs_length in *. 
    lets Hlperm : permut_len (output_async_barbed _ _ Hab_p_q).
    nat_math.
  }}

  lets (X & _) : find_fresh (@nil process).
  lets (p' & Htau_p' & Hcgr_p') : renaming_nTau X Hnd Hflcp. 
  rewrite outputs_get_outs_length; nat_math.

  lets Hab_cc : async_barbed_context_closed Hab_p_q (CParL CHole (wfp_prod_abs_chan (outputs p) lc X)). simpls. 

  lets (q' & Htau_q' & Hcgr_q') : async_barbed_nTau Hab_cc. apply Htau_p'.

  assert (Hout_p : Forall (fun c => perform_output p' c) lc).
  {{
    rewrite <- Forall_Mem. introv Hmx. 
    unfolds perform_output. rewrite <- outputs_trans_out_eq.
    eapply permut_mem. apply output_async_barbed. symmetry.
    apply congr_async_barbed. eassumption. simpls.
    rewrite outputs_ins, outputs_vars. repeat rewrite app_nil_r.
    gen Hllcp x. clear. gen lc. 
    lets Hspec : get_outs_spec p. 

    induction (get_outs p); introv Hl Hm.
    + apply length0 in Hl. subst. inverts Hm.
    + lets (aa & pa & Heq & Hcgr) : Hspec a. constructor. substs.
      destruct lc. rew_length in *. nat_math.
      simpl rename_chan_list. simpl_rename_chan_proc.
      rewrite outputs_wfp_prod_cons. rewrite Mem_app_or_eq.
      simpls. inverts Hm; [left* | right]. constructor.
      specializes~ IHl H0. introv Hm; apply Hspec. repeat constructor~.
  }}

  cut (exists lc', permut lc lc' /\ 
       q' ≡* wfp_Par (wfp_prod (rename_chan_list (get_outs q) lc'))
                     (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))).
  {{
    introv (lc' & Hperm' & Hcgr).
    exists lc'; splits*.
    etransitivity. symmetry. apply congr_async_barbed. eassumption.
    etransitivity. eassumption. applys~ congr_async_barbed.
  }}



  lets Hcgr_prod_abs : wfp_prod_abs_chan_permut_outputs lc X Hab_p_q.
  rewrite outputs_get_outs_length in Hcgr_prod_abs. specializes Hcgr_prod_abs Hllcp. 
  destruct Hcgr_prod_abs as (lc'' & Hperm'' & Hcgr'').
  apply (wfcgr_par q _ (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) _) in Hcgr''.
  Focus 2. etransitivity. apply ordered_wfcgr.
  unfolds ordered; unfolds get_components.
  etransitivity. Focus 2. apply wfcgr_par_assoc_r.
  etransitivity. Focus 2. apply wfcgr_par_com.
  etransitivity. apply wfp_prod_wfcgr_app.
  apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
 
  eapply nTau_wfcgr in Hcgr''; try eassumption.
  destruct Hcgr'' as (q'' & Htau_q'' & Hcgr_q'').
  
  exists lc''. splits*. etransitivity. eassumption.
  assert (Hout_q'' : Forall (fun c => perform_output q'' c) lc).
  {{
    rewrite <- Forall_Mem in *.
    introv Hm; specializes Hout_p Hm.
    eapply async_barbed_out_barb_preserving.
    apply congr_async_barbed. eassumption.
    eapply async_barbed_out_barb_preserving.
    eassumption. auto.
  }} 


    clear dependent p'. clear dependent q'.
    eapply Forall_permut in Hout_q''; try eassumption.
    eapply Forall_permut in Hflcq; try eassumption.
    apply permut_len in Hperm''. rewrite Hperm'' in *.
    clear dependent lc. clear dependent p.
    rename lc'' into lc, q'' into q'.



  rewrite <- outputs_outs in *. 

  


   gen q X q'. inductions lc; intros. 

  + rew_length in *. inverts Htau_q''. symmetry in Hllcq.
    apply length0 in Hllcq. rewrite Hllcq in *. simpl. 
    etransitivity. apply wfcgr_par_assoc_r.
    apply wfcgr_par. reflexivity. apply wfcgr_nil_r.

  + lets Hspec : get_outs_spec q. rename a into c. 
    destruct (get_outs q). rew_length in *. nat_math.
    lets~ (aw & pw & Heq & _) : Hspec w. constructor. substs.
    rewrite outputs_wfp_prod_cons in *. simpl rename_chan_list.
    simpl_rename_chan_proc. etransitivity. Focus 2.
    apply wfcgr_par. symmetry. apply wfp_prod_wfcgr_cons. reflexivity.
    etransitivity. Focus 2. apply wfcgr_par_assoc_l.
    simpls outputs. rewrite app_cons in *. rewrite app_nil_l in *.
    simpls wfp_prod_abs_chan. clear Hspec.

    

Qed. 

(*
Lemma async_barbed_rename_outs_forward_base : forall p q,
  async_barbed (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
               (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) ->
  forall lc, length lc = length (get_outs p) ->
    (Forall (fun c => c ### p /\ c ### q) lc) -> noDup lc ->
    exists gq, permut gq (get_outs q) /\
    async_barbed (wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                 (wfp_Par (wfp_prod (rename_chan_list gq           lc)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) /\
  (forall p'', ~ {{wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) -- Tau ->> AP p''}}) /\
  (forall p'', ~ {{wfp_Par (wfp_prod (rename_chan_list gq           lc)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) -- Tau ->> AP p''}}). *)

(*
  assert (Hf_out_p : Forall (fun c => c ### (wfp_prod (get_outs p))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (Hm' & _). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; rewrite cfresh_par_iff in Hm'. destruct Hm' as (_ & Hm').
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve.
  }}

  assert (Hf_out_q : Forall (fun c => c ### (wfp_prod (get_outs q))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (_ & Hm'). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; rewrite cfresh_par_iff in Hm'. destruct Hm' as (_ & Hm').
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve.
  }}

  assert (Hf_in_p : Forall (fun c => c ### (wfp_prod (get_ins p))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (Hm' & _). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; rewrite cfresh_par_iff in Hm'. destruct Hm' as (_ & Hm').
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve.
  }}

  assert (Hf_in_q : Forall (fun c => c ### (wfp_prod (get_ins q))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (_ & Hm'). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; rewrite cfresh_par_iff in Hm'. destruct Hm' as (_ & Hm').
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve.
  }}

  assert (Hf_var_p : Forall (fun c => c ### (wfp_prod (get_vars p))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (Hm' & _). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve. 
  }}

  assert (Hf_var_q : Forall (fun c => c ### (wfp_prod (get_vars q))) lc).
  {{
    rewrite <- Forall_Mem in *; introv Hm; specializes Hf Hm.
    destruct Hf as (_ & Hm'). eapply wfcgr_alt_freshCh in Hm';
    try apply ordered_wfcgr. unfolds ordered; unfolds get_components.
    eapply wfcgr_alt_freshCh in Hm'. Focus 2. apply wfp_prod_wfcgr_app.
    simpls; fresh_solve.
  }}

  lets Hspec_p : get_outs_spec p.
  lets Hspec_q : get_outs_spec q. 

  (* Tau base *)
      - introv Ht. inverts Ht; subst_wfp; try solve [inverts H2].
      destruct aq; inverts H3.
      eapply get_ins_get_vars_no_tau. eassumption.
    - introv Ht. inverts Ht; subst_wfp; try solve [inverts H2].
      destruct aq; inverts H3.
      eapply get_ins_get_vars_no_tau. eassumption.

*)

Lemma async_barbed_rename_outs_forward_base : forall lc, noDup lc -> forall p q,
  async_barbed (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
               (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) ->
  Forall (fun c => c ### p /\ c ### q) lc ->
    exists gq, permut gq (get_outs q) /\
    async_barbed (wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                 (wfp_Par (wfp_prod (rename_chan_list gq           lc)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))).
fff.Proof with automate.
  introv Hdup. 
  induction p using (measure_induction 
    (fun (p : wfprocess) => (length lc - length (to_list (from_list (outputs p) \n from_list lc))))).
Qed.
  

(*  lets Heq_outs : get_outs_length_async_barbed Hab.
  simpl_gets. repeat rewrite outs_outs, outs_ins, outs_vars, app_nil_r in *. *)
  
  gen lc p q; intro lc. 
  inductions lc; intros.

  + exists (get_outs q). splits*. 
    repeat rewrite~ <- rename_chan_list_nil_idem. 

  + inverts Hnd. specializes IHlc H0; clear H H0.
    
    
    

assert ((∀ q : wfprocess,
         Mem q l
         → ∃ (a : chan) (p' : wfprocess),
           q = wfp_Send a p'
           ∧ (∃ p'' : wfprocess, p ≡* wfp_Par (wfp_Send a p') p''))).
    {{
      introv Hm. apply Hspec_p. repeat constructor~.
    }}
    specializes IHl H; clear H.
    destruct lc. false. 
    assert (Hlc' : length lc = length l) by (rew_length in *; nat_math); clear Hlc.
    
    lets (aa & pa & Heq & _) : Hspec_p a. constructor. substs. 
    assert (Hab' : async_barbed 
      (wfp_Par (wfp_Send aa pa) (wfp_Par (wfp_prod l)
             (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))))
          (wfp_Par (wfp_prod (get_outs q))
             (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).
    {{
      etransitivity. Focus 2. eassumption.
      etransitivity. apply async_barbed_congr_par_assoc_l.
      applys~ async_barbed_congr_par.
      apply congr_async_barbed. symmetry.
      apply wfp_prod_wfcgr_cons.
    }} clear Hab.

    lets Hout_p : TrOut aa pa.
    apply act1_out with (p2 := wfp_Par (wfp_prod l) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) in Hout_p.
    lets (q' & qa & Htq & Hab_q) : async_barbed_out_1 Hab' Hout_p.
    assert (Hab_q' : async_barbed (wfp_Par (wfp_prod l) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) q').
    {{
      etransitivity. Focus 2. eassumption.
      etransitivity. apply async_barbed_congr_nil_l.
      apply async_barbed_congr_par_com.
    }} clear Hab_q.
    
    assert (S (length (get_outs q')) = length (get_outs q)).
    {{
      inverts Htq... Focus 2.
      inverts H2; false... 
      destruct ap; inverts H3. eapply get_ins_no_out; eassumption.
      destruct aq0; inverts H3. eapply get_vars_no_out; eassumption.
      destruct ap; inverts H3.
      
      simpl_gets. rewrite outs_ins, outs_vars, app_nil_r.
      gen H2. clear; intros.
      lets Hspec_q : get_outs_spec q. gen w.
      induction (get_outs q); intros. simpls. inverts H2.
            
      lets (ab & pb & Heq & _) : Hspec_q a. constructor. substs.
      assert (Ht : exists w', {{wfp_Par (wfp_Send ab pb) (wfp_prod l) -- LabOut aa qa ->> AP w'}} /\ get_outs w = get_outs w').
      gen H2. clear. intros. destruct l. simpls. exists (wfp_Par w wfp_Nil).
      splits*. applys~ act1_out. simpl_gets. rewrite~ app_nil_r.
      rewrite wfp_prod_cons in H2. exists w. splits*. clear H2.
      destruct Ht as (w' & Ht' & Heq).
      inverts Ht'; subst_wfp;
      assert (p = wfp_Send ab pb) by rewrite_wfp_equal; substs; clear H.

      destruct ap; inverts H3. inverts H2...
      rewrite Heq. simpl_gets. rew_length.
      replace (length (get_outs (wfp_prod l)))  with (length l). nat_math.
      clear Heq IHl. clear w. inductions l.
      simpls; simpl_gets; auto.
      lets (ac & pc & Heq & _) : Hspec_q a. repeat constructor~. substs.
      destruct l. simpls. simpl_gets; auto.
      rewrite wfp_prod_cons in *. remember (w :: l) as l'.
      clear Heql' w l. simpl_gets. rew_length. erewrite IHl; auto.
      introv Hm. apply Hspec_q. inverts Hm. constructor. repeat constructor~.

      destruct aq; inverts H3.

      specializes IHl H2. introv Hm. apply Hspec_q. repeat constructor~.
      rewrite Heq. simpl_gets. rew_length. nat_math.
    }}

    assert (Hl' : length l = length (get_outs q')) by
      (rew_length in *; nat_math). clear H.

    specializes IHl Hl' Hlc'.
    etransitivity. eassumption.
    etransitivity. apply congr_async_barbed. apply ordered_wfcgr.
    unfolds ordered; unfolds get_components.
    etransitivity. etransitivity. apply congr_async_barbed. apply wfp_prod_wfcgr_app.
    apply async_barbed_congr_par. reflexivity. apply congr_async_barbed. apply wfp_prod_wfcgr_app.
    etransitivity. apply async_barbed_congr_par_com.
    etransitivity. Focus 2. apply async_barbed_congr_par_assoc_r.
    applys~ async_barbed_congr_par.
    
Lemma get_vars_out_invariant : forall p p' p'' a, {{p -- LabOut a p'' ->> AP p'}} ->
  get_vars p = get_vars p'.          
Proof with automate.
  introv Ht. inductions Ht...

  + destruct ap; inverts x.
    specialize (IHHt w eq_refl).
    simpl_gets. rewrite~ IHHt.

  + destruct aq; inverts x.
    specialize (IHHt w eq_refl).
    simpl_gets. rewrite~ IHHt.
Qed.

Lemma get_ins_out_invariant : forall p p' p'' a, {{p -- LabOut a p'' ->> AP p'}} ->
  get_ins p = get_ins p'.          
Proof with automate.
  introv Ht. inductions Ht...

  + destruct ap; inverts x.
    specialize (IHHt w eq_refl).
    simpl_gets. rewrite~ IHHt.

  + destruct aq; inverts x.
    specialize (IHHt w eq_refl).
    simpl_gets. rewrite~ IHHt.
Qed.

Lemma trans_out_cfresh : forall c p p' p'' a, {{p -- LabOut a p'' ->> AP p'}} -> c ### p -> c ### p'.
Proof with automate.
  introv Ht. inductions Ht; introv Hfp...
    simpls; fresh_solve.

    destruct ap; inverts x.
    specialize (IHHt w eq_refl).
    simpls. fresh_solve. 

    destruct aq; inverts x.
    specialize (IHHt w eq_refl).
    simpls. fresh_solve.   
Qed.

    assert (Hf'_var : Forall (fun c => c ### wfp_prod (get_vars q')) lc).
    {{
      erewrite <- get_vars_out_invariant; try eassumption.
      simpl_gets. rewrite vars_outs, vars_ins, vars_vars.
      repeat rewrite app_nil_l. inverts~ Hf_var_q.
    }}

    assert (Hf'_in : Forall (fun c => c ### wfp_prod (get_ins q')) lc).
    {{
      erewrite <- get_ins_out_invariant; try eassumption.
      simpl_gets. rewrite ins_outs, ins_ins, ins_vars.
      rewrite app_nil_l, app_nil_r. inverts~ Hf_in_q.
    }}

    assert (Hf' : Forall (fun c => c ### q') lc).
    {{
      inverts Hf_out_q. inverts Hf_in_q. inverts Hf_var_q.
      rewrite <- Forall_Mem in *. introv Hm.
      eapply trans_out_cfresh. eassumption.      
      specializes H2 Hm; specializes H4 Hm; specializes H6 Hm. 
      simpl; repeat (rewrite cfresh_par_iff; splits* ).
    }}

    specializes~ IHl Hf'_var; auto.
      inverts~ Hf; rewrite <- Forall_Mem in *.
      introv Hm; specializes H2 Hm; specializes~ Hf' Hm.
      inverts~ Hnd.
      inverts Hf_out_p. rewrite <- Forall_Mem in *.
      introv Hm; specializes H2 Hm. destruct l. fresh_solve.
      rewrite wfp_prod_cons in *. simpls; fresh_solve.
      rewrite <- Forall_Mem in *. introv Hm.
      specializes Hf' Hm. eapply wfcgr_alt_freshCh in Hf'.
      Focus 2. apply ordered_wfcgr. unfolds ordered; unfolds get_components.
      eapply wfcgr_alt_freshCh in Hf'. Focus 2. apply wfp_prod_wfcgr_app.
      simpls. rewrite cfresh_par_iff in Hf'. destruct Hf' as (_ & Hf').
      eapply wfcgr_alt_freshCh in Hf'. Focus 2. apply wfp_prod_wfcgr_app.
      simpls; fresh_solve.
      inverts~ Hf_in_p. inverts~ Hf_var_p.

    destruct IHl as (gq' & Hperm_gq' & IH1 & IH2 & IH3).
    clear Hf'_var Hf'_in Hf'.

    assert (Hperm : permut (wfp_Send aa qa :: gq') (get_outs q)).
    {{
      eapply permut_trans. rewrite <- permut_cons_iff. eassumption.
      gen Htq. clear; intros.
      inverts Htq... 

      Focus 2. false.
      destruct aq; inverts H3.
      inverts H2... destruct ap; inverts H4.
      eapply get_ins_no_out; eassumption.
      destruct aq; inverts H4.
      eapply get_vars_no_out; eassumption.

      destruct ap; inverts H3. simpl_gets.
      rewrite outs_ins, outs_vars. repeat rewrite app_nil_r.
      lets Hspec_q : get_outs_spec q.
      gen aa qa w. induction (get_outs q); intros.      
      simpls. inverts H2.

      lets (ab & pb & Heq & _) : Hspec_q a. constructor. substs.
      assert (Ht : exists w', {{wfp_Par (wfp_Send ab pb) (wfp_prod l) -- LabOut aa qa ->> AP w'}} /\ get_outs w = get_outs w').
      gen H2. clear. intros. destruct l. simpls. exists (wfp_Par w wfp_Nil).
      splits*. applys~ act1_out. simpl_gets. rewrite~ app_nil_r.
      rewrite wfp_prod_cons in H2. exists w. splits*. clear H2.
      destruct Ht as (w' & Ht' & Heq). 
      inverts Ht'; subst_wfp;
      assert (p = wfp_Send ab pb) by rewrite_wfp_equal; substs; clear H.

      rewrite Heq. destruct ap; inverts H3. inverts H2...
      simpl_gets. rewrite app_nil_l. rewrite <- permut_cons_iff.
      gen Hspec_q. clear. inductions l; intros. simpls; simpl_gets; auto.
      lets (ac & pc & Heq & _) : Hspec_q a. repeat constructor~. substs.
      destruct l. simpls; simpl_gets; auto.
      rewrite wfp_prod_cons. remember (w :: l) as l'; simpl_gets.
      rewrite app_cons. rewrite app_nil_l. rewrite <- permut_cons_iff.
      eapply IHl. introv Hm. apply Hspec_q. inverts Hm. constructor. repeat constructor~.

      rewrite Heq. destruct aq; inverts H3. specializes IHl H2.
      introv Hm. apply Hspec_q. repeat constructor~.
      simpl_gets. rewrite app_cons. rewrite app_nil_l.
      eapply permut_trans. Focus 2. rewrite <- permut_cons_iff.
      eassumption. applys~ permut_flip_first_two.
    }}

    assert (Hrcl_nil : inputs (wfp_prod (rename_chan_list l lc)) = nil).
    {{
      gen Hspec_p. clear. rename aa into a1, pa into p1. gen lc a1 p1 p. 
      inductions l; intros. simpls~.
      lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
      destruct lc. specializes IHl (@nil chan). 
      rewrite <- rename_chan_list_nil_idem in *. destruct l.
      simpls~. rewrite wfp_prod_cons. remember (w :: l) as l'. 
      simpls. rewrite app_nil_l.
      eapply IHl. introv Hm. apply Hspec_p. inverts Hm.
      constructor. repeat constructor*.
      specializes IHl lc. simpl rename_chan_list. 
      destruct (rename_chan_list l lc).
      simpls; simpl_rename_chan_proc; simpls~.
      remember (w :: l0) as l'. destruct l'.
      simpls. simpl_rename_chan_proc; simpls~.
      rewrite wfp_prod_cons. remember (w0 :: l') as l''.
      simpls. simpl_rename_chan_proc; rewrite app_nil_l.
      eapply IHl. introv Hm. apply Hspec_p. inverts Hm.
      constructor. repeat constructor*.
    }}

    assert (Hrgq'_nil : inputs (wfp_prod (rename_chan_list gq' lc)) = nil).
    {{
      assert (Hspec_gq' : forall q : wfprocess,
            Mem q gq'
            → ∃ (a : chan) (p' : wfprocess),
              q = wfp_Send a p').
      {{
        introv Hm. assert (Mem q0 (get_outs q)). 
        {{
          eapply permut_mem. eassumption. repeat constructor~.
        }}
        lets (b & r & Heq & _) : (get_outs_spec q) H.
        subst. repeat eexists.
      }}
      
      gen Hspec_gq'. clear. gen lc. 
      inductions gq'; intros. simpls~.
      lets (aa & pa & Heq) : Hspec_gq' a. repeat constructor~. subst.
      destruct lc. specializes IHgq' (@nil chan). 
      rewrite <- rename_chan_list_nil_idem in *. destruct gq'.
      simpls~. rewrite wfp_prod_cons. remember (w :: gq') as l'. 
      simpls. rewrite app_nil_l.
      eapply IHgq'. introv Hm. apply Hspec_gq'. repeat constructor~. 
      specializes IHgq' lc. simpl rename_chan_list. 
      destruct (rename_chan_list gq' lc).
      simpls; simpl_rename_chan_proc; simpls~.
      remember (w :: l) as l'. destruct l'.
      simpls. simpl_rename_chan_proc; simpls~.
      rewrite wfp_prod_cons. remember (w0 :: l') as l''.
      simpls. simpl_rename_chan_proc; rewrite app_nil_l.
      eapply IHgq'. introv Hm. apply Hspec_gq'. repeat constructor*.
    }}

    assert (Heq_ins : get_ins q = get_ins q').
    {{
      rewrite <- (get_ins_out_invariant _ _ _ _ Htq).       
      simpl_gets. rewrite ins_outs, ins_ins, ins_vars.
      rewrite app_nil_l, app_nil_r. auto.
    }}

    assert (Heq_vars : get_vars q = get_vars q').
    {{
      rewrite <- (get_vars_out_invariant _ _ _ _ Htq).       
      simpl_gets. rewrite vars_outs, vars_ins, vars_vars.
      repeat rewrite~ app_nil_l. 
    }}

    eexists; splits. eassumption.

    - simpl rename_chan_list. simpl_rename_chan_proc.
      rewrite Heq_ins, Heq_vars.

      etransitivity. apply async_barbed_congr_par. 
      apply congr_async_barbed. apply wfp_prod_wfcgr_cons. reflexivity.
      etransitivity. apply async_barbed_congr_par_assoc_r.

      etransitivity. Focus 2. apply async_barbed_congr_par.
      apply congr_async_barbed. symmetry. apply wfp_prod_wfcgr_cons. reflexivity.
      etransitivity. Focus 2. apply async_barbed_congr_par_assoc_l.

      apply~ async_barbed_congr_par.

      (* WHAT NOW? *)
      admit.

    - simpl rename_chan_list. simpl_rename_chan_proc. 
      cut (forall p'' : wfprocess, ~{{wfp_Par (wfp_Par (wfp_Send c pa) (wfp_prod (rename_chan_list l lc)))
                                             (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) -- Tau ->> AP p''}}).
      {{
        introv Hyp. introv Htau. destruct (rename_chan_list l lc).
        simpls. inverts Htau; subst_wfp;
        assert (p0 = wfp_Send c pa) by rewrite_wfp_equal; substs; clear H.
        inverts H2. destruct aq; inverts H3. eapply get_ins_get_vars_no_tau. eassumption.
        eapply Hyp. eapply TrTau1. apply act1_out. eassumption. eassumption. inverts H2.
        rewrite wfp_prod_cons in *. eapply Hyp. eassumption.
      }}      

      rewrite <- no_tau_channels_disjoint in *. 
      
      simpls. rewrite app_cons; repeat rewrite app_nil_l.
      symmetry. apply fset_extens; introv Hm.
      rewrite in_empty in Hm; false~.
      rewrite <- IH2. rewrite in_inter in *.
      inverts Hm as (Hm1 & Hm2). splits*.
      rewrite from_list_spec in *. repeat rewrite app_cons in *.
      repeat rewrite app_assoc in *. inverts H1. 
      false. rewrite Mem_app_or_eq in H0. inverts H0.
       
      rewrite Hrcl_nil in H. inverts H.
        
      rewrite Mem_app_or_eq in H. inverts H.
      inverts Hf_in_p. rewrite inputs_trans_in_eq in H0.
      destruct H0 as (fp & Ht).
      eapply false_trans_in. apply H2. apply Ht.
      rewrite inputs_vars in H0. inverts H0. auto.

    - simpl rename_chan_list. simpl_rename_chan_proc. 
      cut (forall p'' : wfprocess, ~{{wfp_Par (wfp_Par (wfp_Send c qa) (wfp_prod (rename_chan_list gq' lc)))
                                             (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) -- Tau ->> AP p''}}).
      {{
        introv Hyp. introv Htau. destruct (rename_chan_list gq' lc).
        simpls. inverts Htau; subst_wfp;
        assert (p0 = wfp_Send c qa) by rewrite_wfp_equal; substs; clear H.
        inverts H2. destruct aq; inverts H3. eapply get_ins_get_vars_no_tau. eassumption.
        eapply Hyp. eapply TrTau1. apply act1_out. eassumption. eassumption. inverts H2.
        rewrite wfp_prod_cons in *. eapply Hyp. eassumption.
      }}      

      rewrite <- no_tau_channels_disjoint in *. 
      
      simpls. rewrite app_cons; repeat rewrite app_nil_l.
      symmetry. apply fset_extens; introv Hm.
      rewrite in_empty in Hm; false~.
      rewrite <- IH3. rewrite in_inter in *.
      inverts Hm as (Hm1 & Hm2). splits*.

      rewrite from_list_spec in *. repeat rewrite app_cons in *.
      repeat rewrite app_assoc in *. inverts H1. 
      false. rewrite Mem_app_or_eq in H0. inverts H0.    
      rewrite Hrgq'_nil in H. inverts H.        
      rewrite Mem_app_or_eq in H. inverts H.
      inverts Hf_in_q. rewrite inputs_trans_in_eq in H0.
      destruct H0 as (fp & Ht).
      eapply false_trans_in. apply H2. apply Ht.
      rewrite inputs_vars in H0. inverts H0. 
      erewrite <- get_ins_out_invariant; try eassumption.
      simpl_gets. rewrite ins_outs, ins_ins, ins_vars in *.
      rewrite app_nil_l, app_nil_r. rewrite inputs_vars in *. auto.
      
      rewrite outputs_vars, outputs_ins, app_nil_r in *.
      rewrite from_list_spec in *. inverts~ H1.
      false. rewrite Hrgq'_nil in H0. rewrite app_nil_l in *.
      rewrite inputs_vars in *. rewrite app_nil_r in *.
      rewrite inputs_trans_in_eq in H0. inverts Hf_in_q.
      destruct H0 as (fp & Ht).
      eapply false_trans_in. apply H2. apply Ht.     
Qed.








(* Final Lemmas *)

Lemma async_barbed_in_normal_early_base : forall m p q a fop, 
  {{wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)) -- LabIn a ->> AA fop}} ->
    async_barbed (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) q ->
      exists fq,
        {{q -- LabIn a ->> AA fq}} /\
         async_barbed (wfp_inst_Abs fop (wfp_Send m wfp_Nil)) (wfp_inst_Abs fq (wfp_Send m wfp_Nil)).
Proof with automate.

  assert (forall p, p ≡* wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))).
  {{    
    clear; introv. transitivity (ordered p).
    + apply ordered_wfcgr.
    + unfold ordered; unfold get_components.
      etransitivity. apply wfp_prod_wfcgr_app.
      etransitivity. apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
      etransitivity. apply wfcgr_par_com. apply wfcgr_par_assoc_r.
  }}

  introv Hin_p Hab'.

  cut (∃ foq : abstraction,
   {{wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) -- LabIn a ->> AA foq}}
   ∧ async_barbed (wfp_inst_Abs fop (wfp_Send m wfp_Nil))
       (wfp_inst_Abs foq (wfp_Send m wfp_Nil))).

  introv (foq & Htoq & Hroq).
  lets Hord_q : wfcgr_tr_in q Htoq. symmetry. apply H. 
  destruct Hord_q as (fq & Htq & Hrq). eexists; splits*. 
  etransitivity. eassumption. 
  etransitivity. apply congr_async_barbed. apply Hrq. reflexivity. 
  assert (Hab : async_barbed (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) 
                             (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).
  {{
    etransitivity. eassumption. applys* congr_async_barbed.
  }} clear Hab'.

  assert (Heq_oq : get_outs q = nil).
  {{
    lets Hspec_oq : get_outs_spec q.
    destruct (get_outs q); auto. false.
    lets (aw & pw & Heq & _) : Hspec_oq w. constructor. substs.
    clear Hspec_oq. clear dependent m. 
    symmetry in Hab. apply async_barbed_out_barb_preserving in Hab.
    specializes Hab aw.

    assert (perform_output
              (wfp_Par (wfp_prod (wfp_Send aw pw :: l))
                 (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) aw).
    {{
      destruct l; repeat eexists. 
      + simpl. apply act1_out. constructor.
      + rewrite wfp_prod_cons. repeat apply act1_out. constructor.
    }}

    apply Hab in H0; clear Hab.
    destruct H0 as (p' & p'' & Htp). inverts Htp...
      + destruct ap; inverts H4. eapply get_ins_no_out; eassumption. 
      + destruct aq; inverts H4. eapply get_vars_no_out; eassumption. 
  }}

  rewrite Heq_oq in *; simpl (wfp_prod nil) in *. clear Heq_oq.  
    
  (* Construct the context *)
  set (Ctx := CParR (wfp_Send a (wfp_Send m wfp_Nil)) CHole).
  lets Hcc : async_barbed_context_closed Hab Ctx. simpls.
  clear dependent Ctx. clear H.

  lets Hout_p : TrOut a (wfp_Send m wfp_Nil).
  lets Htau_p : TrTau1 Hout_p Hin_p.

  lets (q' & Htau_q & Hab') : async_barbed_tau Hcc Htau_p.

    inverts Htau_q; 
    assert (p0 = wfp_Send a (wfp_Send m wfp_Nil)) by rewrite_wfp_equal;
    assert (q0 = wfp_Par wfp_Nil (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) by rewrite_wfp_equal;
    subst p0 q0; try solve [inverts H2].

    - clear H H1. destruct aq; inverts H3. inverts H2; subst_wfp; try solve [inverts H3]...
      destruct aq; inverts H4. false; eapply get_ins_get_vars_no_tau; eassumption.
    - clear H H0. inverts H2... assert (p'' = wfp_Send m wfp_Nil) by rewrite_wfp_equal. substs.
      exists fq; splits*.
      etransitivity. apply async_barbed_congr_nil_l.
      etransitivity. apply async_barbed_congr_par_com.
      etransitivity; try eassumption.
      etransitivity. apply async_barbed_congr_par_com.
      apply async_barbed_congr_nil_r.
Qed. 

Lemma async_barbed_in_normal_early : in_normal_relation_early async_barbed.
Proof with automate.


  introv Hab_p_q Hin_p Hfmp Hfmq.

  (* Switching to ordered *)

  assert (forall p, p ≡* wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))).
  {{    
    clear; introv. transitivity (ordered p).
    + apply ordered_wfcgr.
    + unfold ordered; unfold get_components.
      etransitivity. apply wfp_prod_wfcgr_app.
      etransitivity. apply wfcgr_par. reflexivity. apply wfp_prod_wfcgr_app.
      etransitivity. apply wfcgr_par_com. apply wfcgr_par_assoc_r.
  }}

  lets* Hord_p : wfcgr_tr_in (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) Hin_p. 
  destruct Hord_p as (fop & Htop & Hrop).
  assert (Hab_op_oq : async_barbed (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) 
                                   (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).
    etransitivity. apply congr_async_barbed. symmetry. apply H.
    etransitivity. Focus 2. apply congr_async_barbed. apply H. auto.
  assert (Hfmop : m ### (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))).
    eapply wfcgr_alt_freshCh. apply H. auto.
  assert (Hfmoq : m ### (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))).
    eapply wfcgr_alt_freshCh. apply H. auto.
  clear Hab_p_q Hin_p Hfmp Hfmq.

  cut (∃ foq : abstraction,
   {{wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) -- LabIn a ->> AA foq}}
   ∧ async_barbed (wfp_inst_Abs fop (wfp_Send m wfp_Nil))
       (wfp_inst_Abs foq (wfp_Send m wfp_Nil))).

  introv (foq & Htoq & Hroq).
  lets Hord_q : wfcgr_tr_in q Htoq. symmetry. apply H. 
  destruct Hord_q as (fq & Htq & Hrq). eexists; splits*. 
  etransitivity. apply congr_async_barbed. apply Hrop.
  etransitivity. eassumption. 
  etransitivity. apply congr_async_barbed. apply Hrq.
  reflexivity. clear Hrop H fp.

  lets Hspec_op : get_outs_spec p.

  (* 
  remember (get_outs p) as op. clear Heqop.
  destruct op. simpls.

  + assert (Heq_oq : get_outs q = nil).
    {{
      lets Hspec_oq : get_outs_spec q.
      destruct (get_outs q); auto. false.
      lets (aw & pw & Heq & _) : Hspec_oq w. constructor. substs.
      clear Hspec_oq. clear dependent m. 
      symmetry in Hab_op_oq. apply async_barbed_out_barb_preserving in Hab_op_oq.
      specializes Hab_op_oq aw.

      assert (perform_output
                (wfp_Par (wfp_prod (wfp_Send aw pw :: l))
                   (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) aw).
      {{
        destruct l; repeat eexists. 
        + simpl. apply act1_out. constructor.
        + rewrite wfp_prod_cons. repeat apply act1_out. constructor.
      }}

      apply Hab_op_oq in H. clear Hab_op_oq.
      destruct H as (p' & p'' & Htp). inverts Htp...
      + inverts H2.
      + destruct aq; inverts H3. inverts H2...
        - destruct ap; inverts H4. eapply get_ins_no_out; eassumption. 
        - destruct aq; inverts H4. eapply get_vars_no_out; eassumption. 
    }}
    
    rewrite Heq_oq in *; simpl (wfp_prod nil) in *. clear Heq_oq.

    inverts Htop... inverts H2.
    destruct aq; inverts H3. rename a0 into fop; simpl.
    assert (Hab_op_oq' : 
              async_barbed (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))
                           (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))).
    {{
      etransitivity. apply async_barbed_congr_nil_l.
      etransitivity. apply async_barbed_congr_par_com.
      etransitivity. eassumption.
      etransitivity. apply async_barbed_congr_par_com.
      apply async_barbed_congr_nil_r.
    }}

    lets Hyp : async_barbed_in_normal_early_base m H2 Hab_op_oq'.
    destruct Hyp as (foq & Htoq & Hab).
    eexists; splits*. eapply act2_in. eassumption.
    etransitivity. apply async_barbed_congr_par_com.
    etransitivity. apply async_barbed_congr_nil_r.
    etransitivity. eassumption. simpls.
    etransitivity. Focus 2. apply async_barbed_congr_par_com.
    apply async_barbed_congr_nil_l.

    (* There exists an output *) *)
    


Lemma async_barbed_rename_outs_forward : forall p q,
  async_barbed (wfp_Par (wfp_prod (get_outs p)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
               (wfp_Par (wfp_prod (get_outs q)) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) ->
  forall lc, length lc = length (get_outs p) ->
    (Forall (fun c => c ### p /\ c ### q) lc) -> noDup lc ->
    exists lc', permut lc lc' /\
    async_barbed (wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
                 (wfp_Par (wfp_prod (rename_chan_list (get_outs q) lc')) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) /\
  (forall p'', ~ {{wfp_Par (wfp_prod (rename_chan_list (get_outs p) lc)) (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) -- Tau ->> AP p''}}) /\
  (forall p'', ~ {{wfp_Par (wfp_prod (rename_chan_list (get_outs q) lc')) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) -- Tau ->> AP p''}}).
Proof with automate.
  admit.

(*


  induction (get_outs p); intros;
  remember (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) as P;
  remember (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) as Q.

  + apply length_zero_inv in Hlc. subst. clear Hf. simpls.
    
    assert (Heq_oq : get_outs q = nil).
    {{
      lets Hspec_oq : get_outs_spec q.
      destruct (get_outs q); auto. false.
      lets (aw & pw & Heq & _) : Hspec_oq w. constructor. substs.
      clear Hspec_oq.  
      symmetry in Hab. apply async_barbed_out_barb_preserving in Hab.
      specializes Hab aw.

      assert (perform_output
                (wfp_Par (wfp_prod (wfp_Send aw pw :: l))
                   (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) aw).
      {{
        destruct l; repeat eexists. 
        + simpl. apply act1_out. constructor.
        + rewrite wfp_prod_cons. repeat apply act1_out. constructor.
      }}

      apply Hab in H; clear Hab.
      destruct H as (p' & p'' & Htp). inverts Htp...
        + destruct ap; inverts H3. inverts H2. 
        + inverts H2...
          - destruct ap; inverts H3; eapply get_ins_no_out; eassumption. 
          - destruct aq0; inverts H3; eapply get_vars_no_out; eassumption. 
    }} rewrite Heq_oq in *; clear Heq_oq; simpls; splits*.

    - introv Ht. inverts Ht; subst_wfp; try solve [inverts H2].
      destruct aq; inverts H3.
      eapply get_ins_get_vars_no_tau. eassumption.
    - introv Ht. inverts Ht; subst_wfp; try solve [inverts H2].
      destruct aq; inverts H3.
      eapply get_ins_get_vars_no_tau. eassumption.

  + destruct lc. rew_length in *; false. destruct l. 
    - clear IHl. assert (lc = nil).
      apply length_zero_inv. rew_length in Hlc; nat_math. subst. clear Hlc.
      lets (ap & pp & Heq & _) : Hspec_p a. constructor. subst.
      lets Hspec_q : get_outs_spec q. 
      assert (exists pq, get_outs q = wfp_Send ap pq :: nil).
      {{
        assert (Hbarb : perform_output (wfp_Par (wfp_Send ap pp)
                   (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) ap).
          repeat eexists. apply act1_out. constructor.
        eapply async_barbed_out_barb_preserving in Hbarb; try eassumption.

        lets Hspec_oq : get_outs_spec q.
        remember (get_outs q) as oq. clear Heqoq.
        destruct oq. 

        + false. destruct Hbarb as (q' & q'' & Htq). inverts Htq...
          inverts H2. destruct aq; inverts H3.
          inverts H2... destruct ap0; inverts H4.
          eapply get_ins_no_out; eassumption.
          destruct aq; inverts H4.
          eapply get_vars_no_out; eassumption.

        + destruct oq.
          - simpl (wfp_prod (w :: nil)) in *.
            lets (bw & qw & Heq & _) : Hspec_oq w. constructor.
            subst. clear Hspec_oq.
            apply output_async_barbed in Hab.
            simpls. repeat rewrite get_ins_output_empty, get_vars_output_empty in *.
            repeat rewrite app_nil_r in *. exists qw. fequals. fequals.
            apply permut_rem1_lr with (x := ap) in Hab.
            simpls; repeat cases_if*. apply permut_sym in Hab.
            apply permut_of_nil in Hab. inverts Hab.
          - rewrite wfp_prod_cons in *. apply output_async_barbed in Hab.
            remember (wfp_prod (w0 :: oq)) as r. simpl in Hab. 
            repeat rewrite get_ins_output_empty, get_vars_output_empty in *.
            repeat rewrite app_nil_r in *. subst r.
            lets (bw & qw & Heq & _) : Hspec_oq w. constructor.
            lets (bw0 & qw0 & Heq0 & _) : Hspec_oq w0. repeat constructor.
            subst. false. apply permut_len in Hab. 
            destruct oq; simpls; rew_length in *; nat_math.
      }} destruct H as (pq & Heq). rewrite Heq in *. clear Heq Hspec_q Hspec_p.
      
      simpls; simpl_rename_chan_proc. rewrite Forall_to_conj_1 in Hf.
      destruct Hf as (Hfcp & Hfcq). 
      splits. 

      * clear Hnd. 
        lets (X & Hfx) : find_fresh (proc p :: proc q :: nil).
        rewrite Forall_to_conj_2 in Hfx. destruct Hfx as (Hfxp & Hfxq).
        set (Ctx := CParR (wfp_Abs ap X (wfp_Send c (wfp_Gvar X))) CHole).
        lets Hab_cc : async_barbed_context_closed Hab Ctx. simpls.

        lets Hout_p : TrOut ap pp.
        apply act1_out with (p2 := (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) in Hout_p.
        lets Hin_p : TrIn ap X (wfp_Send c (wfp_Gvar X)).
        lets Htau_p : TrTau2 Hin_p Hout_p. simpls... cases_if*. 
        clear e Hout_p Hin_p.
        
        lets (q' & Htau_q & Hab') : async_barbed_tau Hab_cc Htau_p.
        {{
          inverts Htau_q; subst_wfp; 
          assert (p0 = wfp_Abs ap X (wfp_Send c (wfp_Gvar X))) by rewrite_wfp_equal;
          assert (q0 = wfp_Par (wfp_Send ap pq) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) by rewrite_wfp_equal; substs; try solve [inverts H2].

          + clear H1 H. destruct aq; inverts H3. false.
            inverts H2; subst_wfp; assert (p0 = wfp_Send ap pq) by rewrite_wfp_equal; substs. 
            - inverts H3.

            - destruct aq; inverts H4. clear H.
              eapply get_ins_get_vars_no_tau. eassumption.
              
            - inverts H3... clear H.
              apply async_barbed_out_barb_preserving in Hab'.
              specializes Hab' c.
              assert (perform_output
                (wfp_Par (wfp_Send c pp)
                         (wfp_Par wfp_Nil
                                  (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))) c).
              {{
                eexists. eexists. apply act1_out. constructor.
              }}
              specializes Hab' H.
              destruct Hab' as (q' & q'' & Htq).
              inverts Htq; assert (p0 = wfp_Abs a X (wfp_Send c (wfp_Gvar X))) by rewrite_wfp_equal; 
              assert (q0 = wfp_Par wfp_Nil (wfp_inst_Abs fq pq)) by rewrite_wfp_equal; substs.
              inverts H3. 
              clear H2 H0. destruct aq; inverts H5.
              inverts H3... inverts H5.
              destruct aq; inverts H6.
              rewrite Forall_to_conj_1 in *.
              eapply false_trans_out. Focus 2. apply H5.
              eapply wfp_cfresh_in. Focus 2. apply H4.
              simpls; rewrite cfresh_par_iff. splits*.
              rewrite cfresh_send in Hf_out_q; jauto.
              
            - inverts H3.

          + clear H0 H. 
            assert (wfp_inst_Abs fp q'' = wfp_Send c q'').
            {{
              lets Heq : wfp_trans_in_chan_eq H2. substs.
              lets Hyp : wfp_trans_in_abs_eq H2; inverts Hyp as Hyp.
              simpls... cases_if*.
              destruct Hyp as (Y & HneqY & HfY & Heq). substs.
              simpls. rewrite~ <- wfp_subst_decomposition_g... cases_if*. 
            }}
            rewrite H in *. clear H H2. inverts H3;
            assert (p0 = wfp_Send ap pq) by rewrite_wfp_equal; 
            assert (q0 = wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))) by rewrite_wfp_equal; substs; clear H1 H.
            - destruct ap0; inverts H4. inverts H2...
              etransitivity. apply async_barbed_congr_par.
              reflexivity. etransitivity. apply async_barbed_congr_nil_l.
              apply async_barbed_congr_par_com.
              etransitivity. eassumption.
              applys~ async_barbed_congr_par. etransitivity.
              apply async_barbed_congr_par_com.
              apply async_barbed_congr_nil_r.
            - destruct aq; inverts H4.
              false; inverts H2...
              * destruct ap0; inverts H4.
                eapply get_ins_no_out; eassumption.
              * destruct aq; inverts H4.
                eapply get_vars_no_out; eassumption.              
        }}
         
      * rewrite <- no_tau_channels_disjoint. simpl.
        rewrite app_nil_l.



        rewrite inputs_vars, outputs_ins, outputs_vars.
        repeat rewrite app_nil_r.
        symmetry. apply fset_extens; introv Hin. 
        rewrite in_empty in Hin. false.
        rewrite in_inter in Hin. repeat rewrite from_list_spec in *.
        destruct Hin as (Hm1 & Hm2). inverts Hm2; [ | inverts H0].
        rewrite inputs_trans_in_eq in Hm1.
        destruct Hm1 as (fp & Htp). false.
        eapply false_trans_in. Focus 2. eassumption.
        eapply wfcgr_alt_freshCh in Hfcp. Focus 2. apply ordered_wfcgr.
        unfolds ordered; unfolds get_components.
        eapply wfcgr_alt_freshCh in Hfcp. Focus 2. apply wfp_prod_wfcgr_app.
        simpls; rewrite cfresh_par_iff in Hfcp. destruct Hfcp as (_ & Hfcp).
        eapply wfcgr_alt_freshCh in Hfcp. Focus 2. apply wfp_prod_wfcgr_app.
        simpls; fresh_solve.

      * rewrite <- no_tau_channels_disjoint. simpl.
        rewrite app_nil_l.
        rewrite inputs_vars, outputs_ins, outputs_vars.
        repeat rewrite app_nil_r.
        symmetry. apply fset_extens; introv Hin. 
        rewrite in_empty in Hin. false.
        rewrite in_inter in Hin. repeat rewrite from_list_spec in *.
        destruct Hin as (Hm1 & Hm2). inverts Hm2; [ | inverts H0].
        rewrite inputs_trans_in_eq in Hm1.
        destruct Hm1 as (fq & Htq). false.
        eapply false_trans_in. Focus 2. eassumption.
        eapply wfcgr_alt_freshCh in Hfcq. Focus 2. apply ordered_wfcgr.
        unfolds ordered; unfolds get_components.
        eapply wfcgr_alt_freshCh in Hfcq. Focus 2. apply wfp_prod_wfcgr_app.
        simpls; rewrite cfresh_par_iff in Hfcq. destruct Hfcq as (_ & Hfcq).
        eapply wfcgr_alt_freshCh in Hfcq. Focus 2. apply wfp_prod_wfcgr_app.
        simpls; fresh_solve.

    - lets (a1 & p1 & Heq & _) : Hspec_p a. constructor. subst.
      lets (a2 & p2 & Heq & _) : Hspec_p w. repeat constructor. subst.
      destruct lc. rew_length in *. false; nat_math.
      rename c into c1, c0 into c2. 

      simpl rename_chan_list at 1 3 4. repeat rewrite wfp_prod_cons in *.
      replace (rename_chan_proc (wfp_Send a2 p2) c2 :: rename_chan_list l lc) with
              (rename_chan_list (wfp_Send a2 p2 :: l) (c2 :: lc)) by auto.
      remember (wfp_Send a2 p2 :: l) as l'; remember (c2 :: lc) as lc'.
      clear Heql'. clear a2 p2 l. clear Heqlc'. clear c2 lc.
      rename l' into l, lc' into lc. simpl_rename_chan_proc.

      lets Hout_p : TrOut a1 p1.
      apply act1_out with (p2 := wfp_prod l) in Hout_p.
      apply act1_out with (p2 := wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))) in Hout_p.
      eapply async_barbed_out_1 in Hout_p; try eassumption.
      destruct Hout_p as (q' & q'' & Htq & Hab').
      
      inverts Htq...
      Focus 2. false. inverts H2... 
      destruct ap; inverts H3. eapply get_ins_no_out. eassumption.
      destruct aq0; inverts H3. eapply get_vars_no_out. eassumption.

      destruct ap; inverts H3.

      assert (exists q1 lq, permut (get_outs q) (wfp_Send a1 q1 :: lq) /\ w ≡* wfp_prod lq).
      {{
        generalize H2. clear. gen q'' w.
        lets Hspec_q : get_outs_spec q.
        induction (get_outs q); intros.
        + simpls. inverts H2.
        + lets (aa & pa & Heq & Hcgra) : Hspec_q a. constructor. substs.
          destruct l. 
          - simpls. inverts H2. eexists. eexists. splits.
            apply permut_refl. simpls. reflexivity.
          - repeat rewrite wfp_prod_cons in *. remember (w0 :: l) as l'.
            clear Heql' w0 l. inverts H2; subst_wfp;
            assert (p = wfp_Send aa pa) by rewrite_wfp_equal; substs; clear H.
            * destruct ap; inverts H4. inverts H3...
              exists pa l'; splits*. etransitivity.
              apply wfcgr_par_com. apply wfcgr_nil_r.
            * destruct aq; inverts H4.
              lets (q1 & lq & Hper & Hcgr) : IHl H3.
              introv Hm. apply Hspec_q. repeat constructor~.
              exists q1 (wfp_Send aa pa :: lq). splits*.
              eapply permut_trans. apply permut_cons. eassumption.
              apply permut_flip_first_two. apply permut_refl.
              etransitivity. apply wfcgr_par. reflexivity. eassumption.
              clear. destruct lq. simpls. apply wfcgr_nil_r.
              rewrite wfp_prod_cons. reflexivity.
      }}            

      lets Hspec_q : get_outs_spec q.

      (* Moving to the permuted version of q *)
      destruct H as (q1 & q' & Hper & Hcgr).

      assert (Hspec_q' : forall q0 : wfprocess,
            Mem q0 (wfp_Send a1 q1 :: q')
            -> ∃ (a : chan) (p' : wfprocess),
              q0 = wfp_Send a p' /\  
              (∃ p'' : wfprocess, q ≡* wfp_Par (wfp_Send a p') p'')).
      {{
        introv Hm. eapply Hspec_q. eapply permut_mem. apply permut_sym.
        eassumption. auto.        
      }} clear Hspec_q.

      assert (Hab'' : async_barbed
           (wfp_Par (wfp_prod l)
                    (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
           (wfp_Par (wfp_prod q')
                    (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).
      {{
        etransitivity. etransitivity.
        apply async_barbed_congr_par. etransitivity.
        apply async_barbed_congr_nil_l. apply async_barbed_congr_par_com.
        reflexivity. eassumption.
        apply~ async_barbed_congr_par. apply~ congr_async_barbed.
      }} clear Hab'.

      set (Q := (wfp_Par (wfp_prod q') (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).

      assert (Heq_outs : get_outs Q = q').
      {{
        subst Q. simpl_gets. rewrite outs_ins, outs_vars, app_nil_r.
        gen Hspec_q'. clear. gen q a1 q1. inductions q'; intros.
        + simpls~.
        + lets (aa & pa & Heq & _) : Hspec_q' a. repeat constructor~. substs.
          destruct q'. simpls; simpl_gets. auto.
          rewrite wfp_prod_cons. remember (w :: q') as l. simpl_gets.
          rewrite app_cons. rewrite app_nil_l. fequals.
          eapply IHq'. introv Hm. apply Hspec_q'. inverts Hm.
          constructor. repeat constructor~.
      }}

      assert (Heq_ins : get_ins Q = get_ins q).
      {{
        unfold Q. simpl_gets. rewrite <- Heq_outs.
        rewrite ins_outs, ins_ins, ins_vars, app_nil_l, app_nil_r. auto.
      }}

      assert (Heq_vars : get_vars Q = get_vars q).
      {{
        unfold Q. simpl_gets. rewrite <- Heq_outs.
        rewrite vars_outs, vars_ins, vars_vars, app_nil_l. auto.
      }}

      rewrite <- Heq_outs, <- Heq_ins, <- Heq_vars in Hab''. 
      specializes IHl Hab''. introv Hm. apply Hspec_p. constructor*.
      rewrite Heq_outs, Heq_ins, Heq_vars in *. subst Q. 
      clear Heq_outs Heq_ins Heq_vars.

      assert (Hflc : Forall (fun c => c ### wfp_prod (get_vars q)) lc) by
        inverts* Hf_var_q.
      specializes IHl Hflc.
        rew_length in *; nat_math. inverts Hf. clear H1.
        inverts Hf_in_q; inverts Hf_out_q. clear H1 H5.
        rewrite <- Forall_Mem in *. introv Hm. 
        specializes H3 Hm; specializes H4 Hm; specializes H6 Hm;
        specializes Hflc Hm; splits*.
        
        simpls; repeat rewrite cfresh_par_iff; splits*.
        eapply wfcgr_alt_freshCh in H6. Focus 2. 
        eapply wfp_prod_wfcgr_permut. apply permut_sym. eassumption.
        destruct q'. simpls; fresh_solve.
        rewrite wfp_prod_cons in *. simpls; fresh_solve.
        inverts~ Hnd. inverts~ Hf_out_p.
        rewrite <- Forall_Mem in *; introv Hm.
        specializes H3 Hm. simpls; fresh_solve.
        inverts Hf_out_q. rewrite <- Forall_Mem in *; introv Hm.
        specializes H3 Hm. eapply wfcgr_alt_freshCh in H3. Focus 2. 
        eapply wfp_prod_wfcgr_permut. apply permut_sym. eassumption.
        destruct q'. simpls; fresh_solve.
        rewrite wfp_prod_cons in *. simpls; fresh_solve.
        inverts~ Hf_in_p. inverts~ Hf_in_q. 
        inverts~ Hf_var_p. 

      assert (Hrcl_nil : inputs (wfp_prod (rename_chan_list l lc)) = nil).
      {{
        gen Hspec_p. clear. gen lc a1 p1 p. inductions l; intros. simpls~.
        lets (aa & pa & Heq & _) : Hspec_p a. repeat constructor~. subst.
        destruct lc. specializes IHl (@nil chan). 
        rewrite <- rename_chan_list_nil_idem in *. destruct l.
        simpls~. rewrite wfp_prod_cons. remember (w :: l) as l'. 
        simpls. rewrite app_nil_l.
        eapply IHl. introv Hm. apply Hspec_p. inverts Hm.
        constructor. repeat constructor*.
        specializes IHl lc. simpl rename_chan_list. 
        destruct (rename_chan_list l lc).
        simpls; simpl_rename_chan_proc; simpls~.
        remember (w :: l0) as l'. destruct l'.
        simpls. simpl_rename_chan_proc; simpls~.
        rewrite wfp_prod_cons. remember (w0 :: l') as l''.
        simpls. simpl_rename_chan_proc; rewrite app_nil_l.
        eapply IHl. introv Hm. apply Hspec_p. inverts Hm.
        constructor. repeat constructor*.
      }}

      destruct IHl as (IH1 & IH2 & IH3); splits*.

      * clear IH2 IH3. 
        etransitivity. apply async_barbed_congr_par_assoc_r.
        etransitivity. apply async_barbed_congr_par. 
        reflexivity. eassumption.
        etransitivity. apply async_barbed_congr_par_assoc_l.
        applys~ async_barbed_congr_par. 

      * rewrite <- no_tau_channels_disjoint in *. 
        simpls. rewrite app_cons; repeat rewrite app_nil_l.
        symmetry. apply fset_extens; introv Hm.
        rewrite in_empty in Hm; false~.
        rewrite <- IH2. rewrite in_inter in *.
        inverts Hm as (Hm1 & Hm2). splits*.
        rewrite from_list_spec in *. repeat rewrite app_cons in *.
        repeat rewrite app_assoc in *. inverts H1. 
        false. rewrite Mem_app_or_eq in H0. inverts H0.
        
        rewrite Hrcl_nil in H. inverts H.
        
        rewrite Mem_app_or_eq in H. inverts H.
        inverts Hf_in_p. rewrite inputs_trans_in_eq in H0.
        destruct H0 as (fp & Ht).
        eapply false_trans_in. apply H3. apply Ht.
        rewrite inputs_vars in H0. inverts H0. auto.

      * rewrite <- no_tau_channels_disjoint in *. 
        simpls. rewrite app_cons; repeat rewrite app_nil_l.
        symmetry. apply fset_extens; introv Hm.
        rewrite in_empty in Hm; false~.
        rewrite <- IH3. rewrite in_inter in *.
        inverts Hm as (Hm1 & Hm2). splits*.
        rewrite from_list_spec in *. repeat rewrite app_cons in *.
        repeat rewrite app_assoc in *. inverts H1. 
        false. rewrite Mem_app_or_eq in H0. inverts H0.

        rewrite Hrcl_nil in H. inverts H.
        
        rewrite Mem_app_or_eq in H. inverts H.
        inverts Hf_in_q. rewrite inputs_trans_in_eq in H0.
        destruct H0 as (fp & Ht).
        eapply false_trans_in. apply H3. apply Ht.
        rewrite inputs_vars in H0. inverts H0. auto. *)
Qed.  
  
      
(*********** RENAMING ************)

(*
  + assert (Hl : length (w :: op) > 0) by (rew_length in *; nat_math).
    remember (w :: op) as l; clear Heql. *)
    
   














induction op using (measure_induction length); 
  rename H into IH; intros; destruct op.

  + clear Hspec_op IH. 



simpl (wfp_prod nil) in *.
    assert (Heq_oq : get_outs q = nil).
    {{
      lets Hspec_oq : get_outs_spec q.
      destruct (get_outs q); auto. false.
      lets (aw & pw & Heq & _) : Hspec_oq w. constructor. substs.
      clear Hspec_oq. clear dependent m. 
      symmetry in Hab_op_oq. apply async_barbed_out_barb_preserving in Hab_op_oq.
      specializes Hab_op_oq aw.

      assert (perform_output
                (wfp_Par (wfp_prod (wfp_Send aw pw :: l))
                   (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) aw).
      {{
        destruct l; repeat eexists. 
        + simpl. apply act1_out. constructor.
        + rewrite wfp_prod_cons. repeat apply act1_out. constructor.
      }}

      apply Hab_op_oq in H. clear Hab_op_oq.
      destruct H as (p' & p'' & Htp). inverts Htp...
      + inverts H2.
      + destruct aq; inverts H3. inverts H2...
        - destruct ap; inverts H4. eapply get_ins_no_out; eassumption. 
        - destruct aq; inverts H4. eapply get_vars_no_out; eassumption. 
    }}
    
    rewrite Heq_oq in *; simpl (wfp_prod nil) in *. clear Heq_oq.
    
    (* Construct the context *)
    set (Ctx := CParR (wfp_Send a (wfp_Send m wfp_Nil)) CHole).
    assert (Hwf : wf_ctxt Ctx). subst Ctx. repeat constructor*.    
    lets Hcc : async_barbed_context_closed Hab_op_oq Hwf. simpls.
    clear dependent Ctx.

    lets Hout_p : TrOut a (wfp_Send m wfp_Nil).
    lets Htau_p : TrTau1 Hout_p Htop.

    lets (q' & Htau_q & Hab') : async_barbed_tau Hcc Htau_p.

    inverts Htau_q; 
    assert (p0 = wfp_Send a (wfp_Send m wfp_Nil)) by rewrite_wfp_equal;
    assert (q0 = wfp_Par wfp_Nil (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) by rewrite_wfp_equal;
    subst p0 q0.

    - inverts H2.
    - clear H H1. destruct aq; inverts H3. inverts H2; subst_wfp; try solve [inverts H3]...
      destruct aq; inverts H4. false; eapply get_ins_get_vars_no_tau; eassumption.
    - clear H H0. inverts H2... assert (p'' = wfp_Send m wfp_Nil) by rewrite_wfp_equal. substs.
      exists fq; splits*.
      etransitivity. apply async_barbed_congr_nil_l.
      etransitivity. apply async_barbed_congr_par_com.
      etransitivity; try eassumption.
      etransitivity. apply async_barbed_congr_par_com.
      apply async_barbed_congr_nil_r.
    - inverts H2.

  (* Main part - induction step *)

  + 
  {{
    destruct op.

    + simpl (wfp_prod (w :: nil)) in *. clear IH.
      lets (aw & pw & Heq & _) : Hspec_op w. constructor. substs~.

      assert (exists qw, get_outs q = wfp_Send aw qw :: nil).
      {{
        assert (Hbarb : perform_output (wfp_Par (wfp_Send aw pw)
                   (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p)))) aw).
          repeat eexists. apply act1_out. constructor.
        eapply async_barbed_out_barb_preserving in Hbarb; try eassumption.

        lets Hspec_oq : get_outs_spec q.
        remember (get_outs q) as oq. clear Heqoq.
        destruct oq. 

        + false. destruct Hbarb as (q' & q'' & Htq). inverts Htq...
          inverts H2. destruct aq; inverts H3.
          inverts H2... destruct ap; inverts H4.
          eapply get_ins_no_out; eassumption.
          destruct aq; inverts H4.
          eapply get_vars_no_out; eassumption.

        + destruct oq.
          - simpl (wfp_prod (w :: nil)) in *.
            lets (bw & qw & Heq & _) : Hspec_oq w. constructor.
            subst. clear Hspec_oq.
            apply output_async_barbed in Hab_op_oq.
            simpls. repeat rewrite get_ins_output_empty, get_vars_output_empty in *.
            repeat rewrite app_nil_r in *. exists qw. fequals. fequals.
            apply permut_rem1_lr with (x := aw) in Hab_op_oq.
            simpls; repeat cases_if*. apply permut_sym in Hab_op_oq.
            apply permut_of_nil in Hab_op_oq. inverts Hab_op_oq.
          - rewrite wfp_prod_cons in *. apply output_async_barbed in Hab_op_oq.
            remember (wfp_prod (w0 :: oq)) as r. simpl in Hab_op_oq. 
            repeat rewrite get_ins_output_empty, get_vars_output_empty in *.
            repeat rewrite app_nil_r in *. subst r.
            lets (bw & qw & Heq & _) : Hspec_oq w. constructor.
            lets (bw0 & qw0 & Heq0 & _) : Hspec_oq w0. repeat constructor.
            subst. false. apply permut_len in Hab_op_oq. 
            destruct oq; simpls; rew_length in *; nat_math.
      }}

      destruct H as (qw & Heq). rewrite Heq in *. 
      simpl (wfp_prod (_ :: nil)) in *. clear Heq.

      inverts Htop as Htop; subst_wfp;
      assert (p0 = wfp_Send aw pw) by rewrite_wfp_equal; subst. inverts Htop.
      destruct aq; inverts H4; rename a0 into fop. clear H0. simpl.

      
      
      
      lets (c & Hfc) : find_fresh_chan (proc (wfp_Send aw pw) :: proc (wfp_prod (get_ins p)) :: proc (wfp_prod (get_vars p)) :: proc (wfp_Send aw qw) :: proc (wfp_prod (get_ins q)) :: proc (wfp_prod (get_vars q)) :: nil).
      inverts Hfc as Hfc1. inverts H3 as Hfc2. 
      rewrite Forall_to_conj_4 in H3. destruct H3 as (Hfc3 & Hfc4 & Hfc5 & Hfc6).

      lets (X & _) : find_fresh (@nil process).
      
      (* Renaming context *)
      set (Ctx := CParR (wfp_Abs aw X (wfp_Send c (wfp_Gvar X))) CHole).
      assert (Hwf : wf_ctxt Ctx) by repeat constructor.
      lets Hab_op_oq_cc : async_barbed_context_closed Hab_op_oq Hwf.
      subst Ctx. simpls. clear Hwf.

      (* Tau for p *)
      lets Ht_in_p : TrIn aw X (wfp_Send c (wfp_Gvar X)).
      lets Ht_out_p : TrOut aw pw.
      apply act1_out with (p2 := (wfp_Par (wfp_prod (get_ins p))
                            (wfp_prod (get_vars p)))) in Ht_out_p.
      lets Htau_p : TrTau2 Ht_in_p Ht_out_p. simpls... cases_if*. clear e.
      lets (q' & Htau_q & Hab') : async_barbed_tau Hab_op_oq_cc Htau_p.
      
      inverts Htau_q; subst_wfp;
      assert (p0 = wfp_Abs aw X (wfp_Send c (wfp_Gvar X))) by rewrite_wfp_equal;
      assert (q0 = wfp_Par (wfp_Send aw qw) (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q)))) by rewrite_wfp_equal; subst p0 q0; try solve [inverts H2].

      - destruct aq; inverts H3; clear H H1.
        false. apply async_barbed_out_barb_preserving in Hab'.
        specializes Hab' c. apply not_not_intro in Hab'.
        apply Hab'. rew_logic. constructor.
        repeat eexists. apply act1_out. constructor.
        introv (p' & p'' & Htp).
        inverts Htp; subst_wfp; 
        assert (p0 = wfp_Abs aw X (wfp_Send c (wfp_Gvar X))) by rewrite_wfp_equal;
        subst. inverts H3. destruct aq; inverts H4; clear H.
        eapply false_trans_out. Focus 2. exact H3.
        eapply wfp_cfresh_tau. Focus 2. eassumption.
        simpls; repeat (rewrite cfresh_par_iff; splits* ).
    
      - clear H0 H Heq. 
        lets Heq : wfp_trans_in_chan_eq H2; subst a0.
        inverts H3; subst_wfp; 
        assert (p0 = wfp_Send aw qw) by rewrite_wfp_equal; subst.
        * destruct ap; inverts H5. inverts H4... clear H.
          assert (wfp_inst_Abs fp qw = wfp_Send c qw).
            lets Hyp : wfp_trans_in_abs_eq H2.
            inverts Hyp as Hyp. simpls... cases_if*.
            destruct Hyp as (Y & HneqY & Hfy & Heq). subst.
            simpls. rewrite~ <- wfp_subst_decomposition_g.
            automate. cases_if*. 
          rewrite H in *. clear H H2.
          {{
            assert (Hab'' : async_barbed
               (wfp_Par (wfp_Send c pw)
                 (wfp_Par (wfp_prod (get_ins p)) (wfp_prod (get_vars p))))
               (wfp_Par (wfp_Send c qw)
                 (wfp_Par (wfp_prod (get_ins q)) (wfp_prod (get_vars q))))).
            {{
              etransitivity. apply congr_async_barbed. apply wfcgr_par.
              reflexivity. etransitivity. apply wfcgr_nil_l.
              apply wfcgr_par_com. 
              etransitivity. eassumption.
              apply congr_async_barbed. apply wfcgr_par. reflexivity.
              etransitivity. apply wfcgr_par_com. apply wfcgr_nil_r.
            }} clear Hab'; rename Hab'' into Hab'.  

            

            admit.
          }}

      * destruct aq; inverts H5. inverts H4...
        false. destruct ap; inverts H6.
        eapply get_ins_no_out; eassumption.
        false. destruct aq; inverts H6.
        eapply get_vars_no_out; eassumption.

    + rewrite wfp_prod_cons in *. remember (w0 :: op) as op'. 
      clear Heqop' op. rename op' into op.

      
  }} 

  admit.
    
Qed.

(***************
 * CONCLUSIONS *
 ***************)

(* Need 'outputs' for this *)

Lemma rename_chan_list_idem : forall p, 
  rename_chan_list (get_outs p) (outputs (wfp_prod (get_outs p))) = get_outs p.
Proof.
  induction p using (measure_induction wfp_size). 
  destruct_wf p; simpls; simpl_rename_chan_proc; simpl_gets; auto.

  rewrite <- H at 3; solve_wfp_size.
  rewrite <- (H p2) at 3; solve_wfp_size.
  clear H.

  lets Hspec1 : (get_outs_spec p1); lets Hspec2 : (get_outs_spec p2).
  induction (get_outs p1). 

  + repeat (simpl; rewrite~ app_nil_l). 
  + lets (c & p' & Heq & _) : Hspec1 a. constructor. subst.
    destruct l. 
    - rewrite app_cons; rewrite app_nil_l in *.
      destruct (get_outs p2).
      * simpls. rewrite~ app_nil_r.
      * rewrite wfp_prod_cons. remember (w :: l) as l'. clear Heql'. simpl.
        simpl_rename_chan_proc. rewrite app_cons. rewrite~ app_nil_l.
    - repeat rewrite app_cons. repeat rewrite wfp_prod_cons.
      rewrite <- app_cons. remember (w :: l) as l'.  clear Heql'. simpl.
      simpl_rename_chan_proc. rewrite app_cons. fequals.
      apply IHl. introv Hm. apply Hspec1. repeat constructor*.
Qed.

Lemma rename_chan_list_invol : forall p lc,
  rename_chan_list (rename_chan_list (get_outs p) lc) (outputs (wfp_prod (get_outs p))) = get_outs p.
Proof.
  induction p using (measure_induction wfp_size); introv.
  destruct_wf p; simpls; simpl_rename_chan_proc; simpl_gets; auto.

  + destruct lc; simpl; simpl_rename_chan_proc; auto.

  + lets (lc1 & lc2 & Heq1 & Heq2) : rename_chan_list_app_split (get_outs p1) (get_outs p2) lc.
    rewrite Heq2. clear dependent lc.
    rewrite <- (H p1) with (lc := lc1) at 3; solve_wfp_size.
    rewrite <- (H p2) with (lc := lc2) at 3; solve_wfp_size.
    clear H.
    
    lets Hspec1 : (get_outs_spec p1); lets Hspec2 : (get_outs_spec p2).
    gen lc1 lc2. induction (get_outs p1); intros.

    - repeat (rewrite app_nil_l; simpls~).
    - lets (c & p' & Heq & _) : Hspec1 a. constructor. subst.
      destruct l. 
      * destruct lc1.
        simpl rename_chan_list at 2. 
        simpl rename_chan_list at 3. simpl_rename_chan_proc.
        repeat (rewrite app_cons; rewrite app_nil_l).
        destruct (get_outs p2).
        simpls. simpl_rename_chan_proc. auto.
        rewrite wfp_prod_cons. remember (w :: l) as l'.
        simpl. simpl_rename_chan_proc. auto.
        simpl rename_chan_list at 2. simpl_rename_chan_proc.
        simpl rename_chan_list at 3. simpl_rename_chan_proc.
        repeat (rewrite app_cons; rewrite app_nil_l).
        destruct (get_outs p2).
        simpls. simpl_rename_chan_proc. auto.
        rewrite wfp_prod_cons. remember (w :: l) as l'.
        simpl. simpl_rename_chan_proc. auto.
      * repeat rewrite app_cons. repeat rewrite wfp_prod_cons.
        rewrite <- app_cons. remember (w :: l) as l'; clear Heql'.
        destruct lc1.
        simpl rename_chan_list at 2. repeat rewrite app_cons. simpl.
        simpl_rename_chan_proc. rewrite app_cons. fequals.
        specializes IHl (@nil chan) lc2.
        introv Hm. apply Hspec1. repeat constructor~.
        rewrite <- rename_chan_list_nil_idem in *. 
        rewrite~ IHl.
        simpl rename_chan_list at 2. simpl_rename_chan_proc. 
        simpl rename_chan_list at 4. simpl_rename_chan_proc. 
        repeat rewrite app_cons. simpl. simpl_rename_chan_proc.
        fequals. rewrite~ IHl.
        introv Hm. apply Hspec1. repeat constructor~.
Qed. *)



